

from nzpyida import IdaDataBase, IdaDataFrame

from nzpyida.ae import NZFunTApply
from nzpyida.ae import NZFunApply
from nzpyida.ae import NZFunGroupedApply



#dsn = "jdbc:netezza://169.63.46.17:5480/weather"
dsn='weather'
idadb = IdaDataBase(dsn, 'admin', 'password')



#query = "select * from weather_new limit 1000"
print(idadb)

idadf = IdaDataFrame(idadb, 'WEATHER')

query = 'select * from weather limit 10000'




df = idadf.ida_query(query)


code_str_host = """def decision_tree_ml_host(self, df):

    from sklearn.model_selection import cross_val_score
    from sklearn.impute import SimpleImputer
    from sklearn.tree import DecisionTreeClassifier
    from sklearn.model_selection import train_test_split

    from sklearn.preprocessing import LabelEncoder
    import numpy as np

    result = df.groupby('LOCATION')
    #result = df.groupby(pd.qcut(df['ID'], q=3))

    # result = idadf.ida_query(query, autocommit=True)

    for name, group in result:
        # print(name)

        # print(group)
        def decision_tree_classifier(df):
            imputed_df = df.copy()
            ds_size = len(imputed_df)
            temp_dict = dict()





            columns = imputed_df.columns

            for column in columns:
                if column=='ID':
                    continue

                if (imputed_df[column].dtype == 'float64' or imputed_df[column].dtype == 'int64'):
                    if imputed_df[column].isnull().sum()==len(imputed_df):
                     imputed_df[column] = imputed_df[column].fillna(0)

                    else :
                     imp = SimpleImputer(missing_values=np.nan, strategy='mean')
                     transformed_column = imp.fit_transform(imputed_df[column].values.reshape(-1, 1))
                     imputed_df[column] = transformed_column


                if (imputed_df[column].dtype == 'object'):
                    # impute missing values for categorical variables
                    imp = SimpleImputer(missing_values=None, strategy='constant', fill_value='missing')
                    imputed_df[column] = imp.fit_transform(imputed_df[column].values.reshape(-1, 1))
                    imputed_df[column] = imputed_df[column].astype(str)
                    le = LabelEncoder()
                    # print(imputed_df[column].unique())

                    le.fit(imputed_df[column])
                    # print(le.classes_)
                    imputed_df[column] = le.transform(imputed_df[column])
                    temp_dict[column] = le

            X = imputed_df.drop(['RISK_MM', 'RAINTOMORROW'], axis=1)
            y = imputed_df['RAINTOMORROW']
            X_train, X_test, y_train, y_test = train_test_split(X,y, test_size = 0.25, random_state=42, stratify=y)
            #X_train_mod = X_train.drop(['RISK_MM'],axis=1)
            #X_test_mod = X_test.drop(['RISK_MM'],axis=1)
            dt = DecisionTreeClassifier(max_depth=5)
            dt.fit(X_train, y_train)

            accuracy = dt.score(X_test, y_test)    
            #print(accuracy)



            pred_df = X_test.copy()


            y_pred= dt.predict(X_test)
            

            pred_df['RAINTOMORROW'] = y_pred
            
            pred_df['DATASET_SIZE'] = ds_size
            pred_df['CLASSIFIER_ACCURACY']=round(accuracy,2)
            
           
            
            
            




            original_columns = pred_df.columns

            for column in original_columns:

             if column in temp_dict:   
               pred_df[column] = temp_dict[column].inverse_transform(pred_df[column])
               #print(pred_df)

            def print_output(x):
                row = [x['ID'], x['RAINTOMORROW'],  x['DATASET_SIZE'], x['CLASSIFIER_ACCURACY']]
                
                self.output(row)


            pred_df.apply(print_output, axis=1)
                      
            return pred_df
            
 

        ml_result = decision_tree_classifier(df=group)

        """





output_signature = {'ID':'int', 'RAINTOMORROW_PRED' :'str',  'DATASET_SIZE':'int', 'CLASSIFIER_ACCURACY':'float'}


import time
start = time.time()

nz_tapply = NZFunTApply(df=idadf, code_str=code_str_host, fun_name='decision_tree_ml_host', parallel=False,  output_signature=output_signature, merge_output_with_df=True)
result_idadf = nz_tapply.get_result()
result_df = result_idadf.as_dataframe()
idadb.drop_table(result_idadf.tablename)
print("\n")
#pd.set_option('display.max_columns', None)
#pd.set_option('display.width', None)
#pd.set_option('display.max_rows', None)
#pd.set_option('display.max_colwidth', -1)
print(result_df)
end = time.time()
print(end - start)
#result = result.as_dataframe()
groups = result_df.groupby("LOCATION")
for name, group in groups:
    print(name + ":" + str(len(group)))


code_str_apply="""
def apply_fun(self, x):
    from math import sqrt
    max_temp = x[1]
    id = x[0]
    fahren_max_temp = (max_temp*1.8)+32
    row = [id, max_temp,  fahren_max_temp]
    self.output(row)"""
output_signature = {'ID':'int', 'RESULT_MAX_TEMP' :'float', 'RESULT_FAHREN_MAX_TEMP' : 'float'}
nz_apply = NZFunApply(df=idadf, code_str= code_str_apply, fun_name="apply_fun", columns=['ID', 'MAXTEMP'], output_signature=output_signature, merge_output_with_df=True)
result = nz_apply.get_result()
result=result.as_dataframe()
print(result)



end = time.time()
print(end - start)


code_str_host_spus="""def decision_tree_ml(self, df):
            from sklearn.model_selection import cross_val_score
            from sklearn.impute import SimpleImputer
            from sklearn.tree import DecisionTreeClassifier
            from sklearn.model_selection import train_test_split

            from sklearn.preprocessing import LabelEncoder
            import numpy as np

    

            # data preparation
            imputed_df = df.copy()
            ds_size = len(imputed_df)
            temp_dict = dict()
      

            columns = imputed_df.columns

            for column in columns:
                if column=='ID':
                    continue

                if (imputed_df[column].dtype == 'float64' or imputed_df[column].dtype == 'int64'):
                  if imputed_df[column].isnull().sum()==len(imputed_df):
                     imputed_df[column] = imputed_df[column].fillna(0)
                      
                  else :
           
                     imp = SimpleImputer(missing_values=np.nan, strategy='mean')
                     transformed_column = imp.fit_transform(imputed_df[column].values.reshape(-1, 1))         
                     imputed_df[column] = transformed_column
                   
                if (imputed_df[column].dtype == 'object'):
                    # impute missing values for categorical variables
                    imp = SimpleImputer(missing_values=None, strategy='constant', fill_value='missing')
                    imputed_df[column] = imp.fit_transform(imputed_df[column].values.reshape(-1, 1))
                    imputed_df[column] = imputed_df[column].astype('str')
                    le = LabelEncoder()
                  
                    le.fit(imputed_df[column])
                    # print(le.classes_)
                    imputed_df[column] = le.transform(imputed_df[column])
                    temp_dict[column] = le
            
            
            
            # Create a decision tree
            dt = DecisionTreeClassifier(max_depth=5)
            X = imputed_df.drop(['RISK_MM', 'RAINTOMORROW'], axis=1)
            y = imputed_df['RAINTOMORROW']
            X_train, X_test, y_train, y_test = train_test_split(X,y, test_size = 0.25, random_state=42, stratify=y)
           
           
            dt.fit(X_train, y_train)
  
            accuracy = dt.score(X_test, y_test)    
            #print(accuracy)
    
 
   
            pred_df = X_test.copy()
            
    
            y_pred= dt.predict(X_test)
    
            pred_df['RAINTOMORROW'] = y_pred
            pred_df['DATASET_SIZE'] = ds_size
            pred_df['CLASSIFIER_ACCURACY']=round(accuracy,2)
   
            
  

            original_columns = pred_df.columns
  
            for column in original_columns:
        
             if column in temp_dict:   
               pred_df[column] = temp_dict[column].inverse_transform(pred_df[column])
               #print(pred_df)
    
            def print_output(x):
                row = [x['ID'], x['RAINTOMORROW'], x['DATASET_SIZE'], x['CLASSIFIER_ACCURACY']]
                self.output(row)
            
                
            pred_df.apply(print_output, axis=1)
          
            

    

       
"""


output_signature = {'ID':'int', 'RAINTOMORROW_PRED' :'str',  'DATASET_SIZE':'int', 'CLASSIFIER_ACCURACY':'float'}
import time
start = time.time()

nz_groupapply = NZFunGroupedApply(df=idadf,  code_str=code_str_host_spus, index='LOCATION', fun_name="decision_tree_ml", output_signature=output_signature, merge_output_with_df=True)
#nz_groupapply = NZFunGroupedApply(df=idadf,  code_str=code_str_host_spus, index='LOCATION', fun_name="decision_tree_ml")
result = nz_groupapply.get_result()

print("Host+ SPUs execution - slicing on user selection -ML function for partitions within slices\n")
print(result)
groups = result.as_dataframe().groupby("LOCATION")
for name, group in groups:
    print (name +":"+str(len(group)))
end = time.time()
print(end - start)



import time
start = time.time()
output_signature = {'ID':'int', 'RAINTOMORROW_PRED' :'str',  'DATASET_SIZE':'int', 'CLASSIFIER_ACCURACY':'float'}
nz_groupapply = NZFunTApply(df=idadf, code_str=code_str_host_spus, fun_name ="decision_tree_ml", parallel=True, output_signature=output_signature)
result = nz_groupapply.get_result()
result = result.as_dataframe()
print("Host +SPUs execution - slicing on a default column- ML function for the entire slices")
print(result)
print("\n")

end = time.time()
print(end - start)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2015, IBM Corp.
# All rights reserved.
#
# Distributed under the terms of the BSD Simplified License.
#
# The full license is in the LICENSE file, distributed with this software.
# -----------------------------------------------------------------------------

"""
In-Database User defined functions
"""
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division
from __future__ import absolute_import

import inspect
import textwrap
from builtins import dict
from future import standard_library
from nzpyida import IdaDataBase, IdaDataFrame

from nzpyida.ae import shaper, result_builder

standard_library.install_aliases()


class NZInstall(object):

    def __init__(self, idadb, package_name):
        """
        Constructor for install
        """
        self.package_name=package_name
        self.db =idadb



    def  getResultCode(self):

        
        ae_name = "nzpy..py_udtf_install"

        output_signature = {'ResultCode': 'int'}
        base_code = shaper.get_base_shaper_install(output_signature, self.package_name)

        run_string = textwrap.dedent(""" BaseShaperUdtf.run()""")

        final_code = base_code + "\n" + run_string

        columns_string = "'CODE_TO_EXECUTE=" + "\"" + final_code + "\"" + "'"
        query = "select * from table with final (" + ae_name + "(" + columns_string + ")) "


        result = self.db.ida_query(query)

        if(len(result.values)>0):
            return result.values[0]









#!/usr/bin/env python
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright (c) 2015, IBM Corp.
# All rights reserved.
#
# Distributed under the terms of the BSD Simplified License.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------

"""
In-Database K-means. Copies the interface of sklearn.cluster.KMeans
"""
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division
from __future__ import absolute_import
from builtins import int
from builtins import dict
from builtins import str
from future import standard_library
standard_library.install_aliases()

from lazy import lazy
import pandas as pd

import nzpyida
from nzpyida import IdaDataFrame
from nzpyida.exceptions import IdaKMeansError
import six

class KMeans(object):
    """
    The K-means algorithm is the most widely used clustering algorithm that 
    uses an explicit distance measure to partition the data set into clusters.

    The K-means algorithm represents each cluster by the vector of the mean 
    attribute values of all training instances - for numeric attributes - and 
    by the vector of modal (most frequent) values - for nominal attributes - 
    that are assigned to that cluster. This cluster representation is called 
    cluster center.


    The KMeans class provides an interface for using the KMEANS
    and PREDICT_KMEANS IDAX/INZA methods.
    """
    def __init__(self, n_clusters=3, modelname=None, max_iter=5, distance="euclidean",
                 random_state=12345, idbased=False, statistics=None):
        """ 
        Constructor for K-means clustering.

        Parameters
        ----------

        n_cluster : int, optional, default: 3
            The number of cluster centers.
            Range : > 2

        modelname : str, optional
            The name of the clustering model that will be built.
            It should contain only alphanumeric characters and underscores.
            All lower case characters will be converted to upper case characters.
            If it is not given, it will be generated automatically.
            If the parameter corresponds to an existing model in the database,
            it is replaced during the fitting step.


        max_iter : int, > 1 and <= 1000, default = 5
            The maximum number of iterations.

        distance : str, default: "euclidean"
             The distance function. The following values are allowed: “euclidean” and “norm_euclidean”.


        random_state : int, default: 12345
            The random seed of the generator.

        idbased : bool, optional, default: False
            Specifies that the random seed of the generator is based on the value of the ID column.

        statistics : str, optional
            Indicates the statistics that are collected.

            The following values are allowed: ‘none’, ‘columns’, ‘values:n’, and ‘all’:
                * If statistics='none' is specified, no statistics are collected.
                * If statistics='columns' is specified, statistics on the columns
                  of the input table are collected, for example, mean values.
                * If statistics='values:n' is specified, and if n is a positive number, statistics
                  on the columns and the column values are collected.

                    Up to <n> column value statistics are collected.
                        * If a nominal column contains more than <n> values,
                          only the <n> most frequent column statistics are kept.
                        * If a numeric column contains more than <n> values, the values
                          are discretized, and the statistics are collected on the discretized values.
                * statistics=all is identical to statistics=values:100.

        Attributes
        ----------
        centers: table containing the coordinates of each cluster center;

        cluster_centers_: coordinates of each cluster center, as list;

        withinss: Within cluster sum of squares, by cluster, as a list;

        size_clusters: number of items in each cluster, as list;

        inertia_: float, total inertia of the system, defined as the sum of each cluster's within cluster sum of squares.

        Returns
        -------
            The KMeans object, ready to be used for fitting and prediction

        Examples
        --------
        >>> idadb = IdaDataBase("BLUDB")
        >>> idadf = IdaDataFrame(idadb, "IRIS", indexer = "ID")
        >>> kmeans = KMeans(3) # clustering with 3 clusters
        >>> kmeans.fit(idadf)
        >>> kmeans.predict(idadf)

        Notes
        -----
        Inner parameters of the model can be printed and modified by using 
        get_params and set_params. But we recommend creating a new KMeans model 
        instead of modifying it.

        """
        self.modelname = modelname
        self.n_clusters = n_clusters
        self.max_iter = max_iter
        self.distance = distance
        self.random_state = random_state
        self.idbased = idbased
        self.statistics = statistics

        # Get set at fit step
        self._idadb = None
        self._idadf = None
        self._column_id = None
        self.incolumn = None
        self.coldeftype = None
        self.coldefrole = None
        self.colPropertiesTable = None

        # Get set at predict step
        self.outtable = None

    @lazy
    def labels_(self):
        """
        Return the corresponding labels for each ID.
        """
        try:
            return self.predict(self._idadf, self._column_id)
        except:
            raise AttributeError(str(self.__class__) + " object has no attribute 'labels_'")

    def get_params(self):
        """
        Return the parameters of the K-means clustering.
        """
        params = dict()
        params['modelname'] = self.modelname
        params['n_clusters'] = self.n_clusters
        params['max_iter'] = self.max_iter
        params['distance'] = self.distance
        params['random_state'] = self.random_state
        params['idbased'] = self.idbased
        params['statistics'] = self.statistics

        params['incolumn'] = self.incolumn
        params['coldeftype'] = self.coldeftype
        params['coldefrole'] = self.coldefrole
        params['colPropertiesTable'] = self.colPropertiesTable

        params['outtable'] = self.outtable
        return params

    def set_params(self, **params):
        """
        Change the parameters of the K-means clustering.
        """
        if not params:
            # Simple optimisation to gain speed (inspect is slow)
            return self
        valid_params = self.get_params()
        for key, value in six.iteritems(params):
            if key not in valid_params:
                raise ValueError('Invalid parameter %s for estimator %s' %
                                     (key, self.__class__.__name__))
            setattr(self, key, value)
        return self

    def fit(self, idadf, column_id="ID", incolumn=None, coldeftype=None,
            coldefrole=None, colPropertiesTable=None, verbose=False):
        """
        Use the KMEANS stored procedure to build a K-means clustering model 
        that clusters the input data into k centers.


        Parameters
        ----------
        idadf : IdaDataFrame
            The name of the input IdaDataFrame.

        column_id : str, default: "ID"
            The column of the input IdaDataFrame that identifies a unique
            instance ID.

        incolumn : dict, optional
            The columns of the input table that have specific properties, which
            are separated by a semi-colon (;).
            Each column is succeeded by one or more of the following properties:

                * By type nominal (':nom') or by type continuous (':cont'). By default,
                  numerical types are continuous, and all other types nominal.
                * By role ':id', ':target', ':input', or ':ignore'.

            If this parameter is not specified, all columns of the input table
            have default properties.

        coldeftype : dict, optional
            The default type of the input table columns. The following values 
            are allowed: ‘nom’ and ‘cont’. If the parameter is not specified, 
            numeric columns are continuous and all other columns are nominal.


        coldefrole : dict, optional
            The default role of the input table columns. The following values 
            are allowed: ‘input’ and ‘ignore’. If the parameter is not 
            specified, all columns are input columns.

        colPropertiesTable : idaDataFrame, optional
            The input IdaDataFrame or name of the table where the properties of the
            columns of the input IdaDataFrame (idadf) are stored.
            The table name should contain only alphanumeric characters and underscores.
            All lower case characters will be converted to upper case characters.
            If this parameter is not specified, the column properties of the input table column
            properties are detected automatically.

        verbose : bool, default: False
            Verbosity mode.

        """
        if not type(idadf).__name__ == 'IdaDataFrame':
            raise TypeError("Argument should be an IdaDataFrame")

        idadf._idadb._check_procedure("KMEANS", "KMeans")

        # Check the ID
        if column_id not in idadf.columns:
            raise ValueError("No id columns is available in IdaDataFrame:" + column_id +
                             ". Either create a new ID column using add_column_id function" +
                             " or give the name of a column that can be used as ID")

        self._idadb = idadf._idadb
        self._idadf = idadf
        self._column_id = column_id

        # Check or create a model name
        if self.modelname is None:
            self.modelname = idadf._idadb._get_valid_modelname('KMEANS_')
        else:
            self.modelname = nzpyida.utils.check_modelname(self.modelname)
            if idadf._idadb.exists_model(self.modelname):
                idadf._idadb.drop_model(self.modelname)

        self.incolumn = incolumn
        self.coldeftype = coldeftype
        self.coldefrole = coldefrole
        if not (colPropertiesTable is None or isinstance(colPropertiesTable, nzpyida.frame.IdaDataFrame)):
            colPropertiesTable = nzpyida.utils.check_tablename(colPropertiesTable)
        self.colPropertiesTable = colPropertiesTable

        # Create a temporay view
        tmp_view_name = idadf.internal_state._create_view()
        # tmp_view_name = idadf.internal_state.current_state # deprecated, hange to idadf.name
        
        #if "." in tmp_view_name:
            #tmp_view_name = tmp_view_name.split('.')[-1]
        if (self.outtable == None) & self._idadb._is_netezza_system():
            self.outtable = self._idadb._get_valid_tablename(prefix="OUTTABLE_KMEANS_")
        try:
            # TODO: outtable is optional but this does not match with the doc
            # Defect to declare
            idadf._idadb._call_stored_procedure("KMEANS ",
                                                model = self.modelname,
                                                intable = tmp_view_name,
                                                k = self.n_clusters,
                                                maxiter = self.max_iter,
                                                outtable = self.outtable,
                                                distance = self.distance,
                                                id = self._column_id,
                                                randseed = self.random_state,
                                                statistics = self.statistics,
                                                idbased = self.idbased,
                                                incolumn = self.incolumn,
                                                coldeftype = self.coldeftype,
                                                coldefrole = self.coldefrole,
                                                colPropertiesTable = self.colPropertiesTable)

        except:
            raise
        finally:
            idadf.internal_state._delete_view()
            idadf.commit()

        result = self._retrieve_KMeans_Model(self.modelname, verbose)
        self.centers = result['centers']
        self.cluster_centers_ = result['centers'].values
        self.withinss = result['withinss']
        self.size_clusters = [int(str(x)) for x in result['size']]
        self.inertia_ = sum(self.withinss)

        if verbose is True:
            self.describe()

        return

    def predict(self, idadf, column_id=None, outtable=None):
        """
        Apply the K-means clustering model to new data.

        Parameters
        ----------
        idadf : IdaDataFrame
            IdaDataFrame to be used as input.

        column_id : str
            The column of the input table that identifies a unique instance ID.
            Default: the same id column that is specified in the stored procedure to build the model.

        outtable : str
            The name of the output table where the assigned clusters are stored.
            It should contain only alphanumeric characters and underscores.
            All lower case characters will be converted to upper case characters.
            If this parameter is not specified, it is generated automatically.
            If the parameter corresponds to an existing table in the database,
            it is replaced.

        Returns
        -------
        IdaDataFrame
            IdaDataFrame containing the closest cluster for each data point referenced by its ID.
        """
        if not type(idadf).__name__ == 'IdaDataFrame':
            raise TypeError("Argument should be an IdaDataFrame")

        # Check the ID
        if column_id is None:
            column_id = self._column_id
        if column_id not in idadf.columns:
            raise ValueError("No id columns is available in IdaDataFrame:" + column_id +
                             ". Either create a new ID column using add_column_id function" +
                             " or give the name of a column that can be used as ID")

        if self._idadb is None:
            raise IdaKMeansError("No KMeans model was trained before")


        if outtable is None:
            outtable = idadf._idadb._get_valid_modelname('PREDICT_KMEANS_')
        else:
            if self.outtable:
                outtable = self.outtable
            outtable = nzpyida.utils.check_tablename(outtable)
            if idadf._idadb.exists_table(outtable):
                idadf._idadb.drop_table(outtable)

        self.outtable = outtable
        # Create a temporay view
        tmp_view_name = idadf.internal_state._create_view()
        # tmp_view_name = idadf.internal_state.current_state
        
        #if "." in tmp_view_name:
            #tmp_view_name = tmp_view_name.split('.')[-1]
            
        try:
            idadf._idadb._call_stored_procedure("PREDICT_KMEANS ",
                                                 model = self.modelname,
                                                 intable = tmp_view_name,
                                                 id = column_id,
                                                 outtable = self.outtable
                                                 )
        except:
            raise
        finally:
            idadf.internal_state._delete_view()
            idadf._idadb.commit()

        self.labels_ = nzpyida.IdaDataFrame(idadf._idadb, outtable, indexer=column_id)
        return self.labels_

    def fit_predict(self, idadf, column_id="ID", incolumn=None, coldeftype=None,
                    coldefrole=None, colPropertiesTable=None, outtable=None,
                    verbose=False):
        """
        Convenience function for fitting the model and using it to make 
        predictions on the same dataset. See the fit and predict documentation 
        for an explanation about their attributes.
        """
        self.fit(idadf, column_id, incolumn, coldeftype, coldefrole,
                 colPropertiesTable, verbose)
        return self.predict(idadf, column_id, outtable)

    def describe(self):
        """
        Return a description of the K-means clustering, if a prediction was 
        made. Otherwise,  this function returns the parameters of the model.
        """
        if self._idadb is None:
            return self.get_params
        else:
            print("KMeans clustering with " + str(self.n_clusters) +
            " clusters of sizes " + ', '.join([str(x) for x in self.size_clusters]))
            print()
            print("Cluster means: ")
            print(self.centers)
            print()
            print("Within cluster sum of squares by cluster:")
            print(self.withinss)
            try:
                self._idadb._call_stored_procedure("PRINT_MODEL ", model = self.modelname)
            except:
                raise
            return

    def _retrieve_KMeans_Model(self, modelname, verbose=False):
        """
        Retrieve information about the model to print the results. The KMEANS 
        IDAX/INZA function stores its result in 4 tables:

            * <MODELNAME>_MODEL
            * <MODELNAME>_COLUMNS
            * <MODELNAME>_COLUMN_STATISTICS
            * <MODELNAME>_CLUSTERS

        Parameters
        ----------
        modelname : str
            The name of the model that is retrieved.

        verbose : bol, default: False
            Verbosity mode.
        """
        modelname = nzpyida.utils.check_modelname(modelname)

        if self._idadb is None:
            raise IdaKMeansError("No KMeans model was trained before")

        if self._idadb._is_netezza_system():
            modeltable_prefix = "INZA.NZA_META_"
        else:
            modeltable_prefix = ""
        model_main = self._idadb.ida_query('SELECT * FROM ' + modeltable_prefix + modelname + '_MODEL')
        # Woraround for specific version of ODBC
        model_main.columns = ['MODELCLASS', 'COMPARISONTYPE', 'COMPARISONMEASURE', 'NUMCLUSTERS']
        model_main.columns = [x.upper() for x in model_main.columns]


        col_info = self._idadb.ida_query('SELECT * FROM ' + modeltable_prefix + modelname + '_COLUMNS',)
        col_info.columns = ['COLUMNNAME', 'DATATYPE', 'OPTYPE', 'USAGETYPE', 'COLUMNWEIGHT',
       'AUTOTRANSFORM', 'TRANSFORMEDCOLUMN', 'COMPAREFUNCTION', 'IMPORTANCE',
       'OUTLIERTREATMENT', 'LOWERLIMIT', 'UPPERLIMIT', 'CLOSURE',
       'STATISTICSTYPE']
        col_info.columns = [x.upper() for x in col_info.columns]

        col_stats = self._idadb.ida_query('SELECT * FROM ' + modeltable_prefix + modelname + '_COLUMN_STATISTICS')
        col_stats.columns = ['CLUSTERID', 'COLUMNNAME', 'CARDINALITY', 'MODE', 'MINIMUM', 'MAXIMUM',
       'MEAN', 'VARIANCE', 'VALIDFREQ', 'MISSINGFREQ', 'INVALIDFREQ',
       'IMPORTANCE']
        col_stats.columns = [x.upper() for x in col_stats.columns]

        km_out_stat = self._idadb.ida_query('SELECT * FROM ' + modeltable_prefix + modelname + '_CLUSTERS')
        km_out_stat.columns = ['CLUSTERID', 'NAME', 'DESCRIPTION', 'SIZE', 'RELSIZE', 'WITHINSS']
        km_out_stat.columns = [x.upper() for x in km_out_stat.columns]

        k = model_main.iloc[0][3]
        distance = model_main.iloc[0][2]
        cont_cols = col_info.loc[(col_info['USAGETYPE'] == 'active') & (col_info['OPTYPE'] == 'continuous'), ['COLUMNNAME']]
        cat_cols = col_info.loc[(col_info['USAGETYPE'] == 'active') & (col_info['OPTYPE'] == 'categorical'), ['COLUMNNAME']]

        columns = []
        for x in col_stats['COLUMNNAME'].values:
            if x not in columns:
                columns.append(x)

        clusters = km_out_stat['CLUSTERID'].values
        clusters.sort()

        cluster_centers = []
        for cluster in clusters:
            tmp = [cluster]
            for column in columns:
                if column in cont_cols.values:
                    tmp.append(col_stats.loc[(col_stats['CLUSTERID'] == cluster) & (col_stats['COLUMNNAME'] == column)]['MEAN'].values[0])
                elif column in cat_cols.values:
                    tmp.append(col_stats.loc[(col_stats['CLUSTERID'] == cluster) & (col_stats['COLUMNNAME'] == column)]['MODE'].values[0])
                else:
                    raise TypeError("Unexpected column category")
            cluster_centers.append(tmp)

        centers = pd.DataFrame([tuple(x) for x in cluster_centers])
        centers.columns = ['CLUSTERID'] + columns

        if verbose is True:
            print("MODEL")
            print(model_main)
            print("COLUMNS")
            print(col_info)
            print("COLUMNS_STATISTICS")
            print(col_stats)
            print("CLUSTERS")
            print(km_out_stat)

        result = dict()
        result['withinss'] = km_out_stat['WITHINSS'].values
        result['size'] = km_out_stat['SIZE'].values
        result['relsize'] = km_out_stat['RELSIZE'].values
        result['distance'] = distance

        result['k'] = k
        result['centers'] = centers

        return result


#!/usr/bin/env python
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright (c) 2015, IBM Corp.
# All rights reserved.
#
# Distributed under the terms of the BSD Simplified License.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------

"""
Test module for IdaGeoSeries
"""
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from future import standard_library
standard_library.install_aliases()

import pandas
import pytest
import six

from nzpyida import IdaSeries
from nzpyida import IdaGeoSeries
from nzpyida.exceptions import IdaGeoDataFrameError

@pytest.mark.skipif("'netezza' in config.getvalue('jdbc') or config.getvalue('hostname') != ''")
class Test_IdaGeoSeries(object):

    def test_idageoseries_generalize(self, idageodf_county):
        idageoseries = idageodf_county['SHAPE']
        with pytest.raises(TypeError):
            idageoseries.generalize('not a float')
        with pytest.raises(ValueError):
            idageoseries.generalize(-5)
        assert(isinstance(idageoseries.generalize(1.0), IdaGeoSeries))

    def test_idageoseries_buffer(self, idageodf_county):
        idageoseries = idageodf_county['SHAPE']
        with pytest.raises(TypeError):
            idageoseries.buffer(distance='not a number')
        assert(isinstance(idageoseries.buffer(
                distance=2.3, unit="CLARKE'S FOOT"), IdaGeoSeries))

    def test_idageoseries_centroid(self, idageodf_county):
        idageoseries = idageodf_county['SHAPE']
        assert(isinstance(idageoseries.centroid(), IdaGeoSeries))

    def test_idageoseries_convex_hull(self, idageodf_county):
        idageoseries = idageodf_county['SHAPE']
        assert(isinstance(idageoseries.convex_hull(), IdaGeoSeries))

    def test_idageoseries_boundary(self, idageodf_county):
        idageoseries = idageodf_county['SHAPE']
        assert(isinstance(idageoseries.boundary(), IdaGeoSeries))

    def test_idageoseries_envelope(self, idageodf_county):
        idageoseries = idageodf_county['SHAPE']
        assert(isinstance(idageoseries.envelope(), IdaGeoSeries))

    def test_idageoseries_exterior_ring(self, idageodf_county):
        idageoseries = idageodf_county['SHAPE'].envelope() # get ST_POLYGON
        assert(isinstance(idageoseries.exterior_ring(), IdaGeoSeries))

    def test_idageoseries_mbr(self, idageodf_county):
        idageoseries = idageodf_county['SHAPE']
        assert(isinstance(idageoseries.mbr(), IdaGeoSeries))

    def test_idageoseries_end_point(self, idageodf_county):
        # TODO: add dataset with a column with ST_LINESTRING
        pass

    def test_idageoseries_mid_point(self, idageodf_county):
        # TODO: add dataset with a column with ST_LINESTRING
        pass

    def test_idageoseries_start_point(self, idageodf_county):
        # TODO: add dataset with a column with ST_LINESTRING
        pass

    def test_idageoseries_srid(self, idageodf_county):
        idageoseries = idageodf_county['SHAPE']
        assert(isinstance(idageoseries.srid(), IdaSeries))

    def test_idageoseries_srs_name(self, idageodf_county):
        idageoseries = idageodf_county['SHAPE']
        assert(isinstance(idageoseries.srs_name(), IdaSeries))

    def test_idageoseries_geometry_type(self, idageodf_county):
        idageoseries = idageodf_county['SHAPE']
        assert(isinstance(idageoseries.geometry_type(), IdaSeries))

    def test_idageoseries_area(self, idageodf_county):
        idageoseries = idageodf_county['SHAPE']
        assert(isinstance(idageoseries.area(
                unit="BRITISH FOOT (BENOIT 1895 A)"), IdaSeries))

    def test_idageoseries_dimension(self, idageodf_county):
        idageoseries = idageodf_county['SHAPE']
        assert(isinstance(idageoseries.dimension(), IdaSeries))  

    def test_idageoseries_length(self, idageodf_tornado):
        idageoseries = idageodf_tornado['SHAPE']
        assert(isinstance(idageoseries.length(
                unit="CLARKE'S FOOT"), IdaSeries))

    def test_idageoseries_perimeter(self, idageodf_county):
        idageoseries = idageodf_county['SHAPE']
        assert(isinstance(idageoseries.perimeter(
                unit="CLARKE'S FOOT"), IdaSeries))

    def test_idageoseries_num_geometries(self, idageodf_county):
        idageoseries = idageodf_county['SHAPE']
        assert(isinstance(idageoseries.num_geometries(), IdaSeries))

    def test_idageoseries_num_interior_ring(self, idageodf_county):
        idageoseries = idageodf_county['SHAPE'].envelope() # get ST_POLYGON
        assert(isinstance(idageoseries.num_interior_ring(), IdaSeries))

    def test_idageoseries_num_line_strings(self, idageodf_tornado):
        idageoseries = idageodf_tornado['SHAPE']
        assert(isinstance(idageoseries.num_line_strings(), IdaSeries))

    def test_idageoseries_num_points(self, idageodf_county):
        idageoseries = idageodf_county['SHAPE']
        assert(isinstance(idageoseries.num_points(), IdaSeries))

    def test_idageoseries_num_polygons(self, idageodf_county):
        idageoseries = idageodf_county['SHAPE']
        assert(isinstance(idageoseries.num_polygons(), IdaSeries))

    def test_idageoseries_coord_dim(self, idageodf_county):
        idageoseries = idageodf_county['SHAPE']
        assert(isinstance(idageoseries.coord_dim(), IdaSeries))

    def test_idageoseries_is_3d(self, idageodf_county):
        idageoseries = idageodf_county['SHAPE']
        assert(isinstance(idageoseries.is_3d(), IdaSeries))

    def test_idageoseries_is_measured(self, idageodf_county):
        idageoseries = idageodf_county['SHAPE']
        assert(isinstance(idageoseries.is_measured(), IdaSeries))

    def test_idageoseries_is_valid(self, idageodf_county):
        idageoseries = idageodf_county['SHAPE']
        assert(isinstance(idageoseries.is_valid(), IdaSeries))
        
    def test_idageoseries_max_m(self, idageodf_county):
        idageoseries = idageodf_county['SHAPE']
        assert(isinstance(idageoseries.max_m(), IdaSeries))

    def test_idageoseries_max_x(self, idageodf_county):
        idageoseries = idageodf_county['SHAPE']
        assert(isinstance(idageoseries.max_x(), IdaSeries))

    def test_idageoseries_max_y(self, idageodf_county):
        idageoseries = idageodf_county['SHAPE']
        assert(isinstance(idageoseries.max_y(), IdaSeries))

    def test_idageoseries_max_z(self, idageodf_county):
        idageoseries = idageodf_county['SHAPE']
        assert(isinstance(idageoseries.max_z(), IdaSeries))
        
    def test_idageoseries_min_m(self, idageodf_county):
        idageoseries = idageodf_county['SHAPE']
        assert(isinstance(idageoseries.min_m(), IdaSeries))

    def test_idageoseries_min_x(self, idageodf_county):
        idageoseries = idageodf_county['SHAPE']
        assert(isinstance(idageoseries.min_x(), IdaSeries))

    def test_idageoseries_min_y(self, idageodf_county):
        idageoseries = idageodf_county['SHAPE']
        assert(isinstance(idageoseries.min_y(), IdaSeries))

    def test_idageoseries_min_z(self, idageodf_county):
        idageoseries = idageodf_county['SHAPE']
        assert(isinstance(idageoseries.min_z(), IdaSeries))

    def test_idageoseries_m(self, idageodf_county):
        idageoseries = idageodf_county['SHAPE'].centroid()
        assert(isinstance(idageoseries.m(), IdaSeries))

    def test_idageoseries_x(self, idageodf_county):
        idageoseries = idageodf_county['SHAPE'].centroid()
        assert(isinstance(idageoseries.x(), IdaSeries))

    def test_idageoseries_y(self, idageodf_county):
        idageoseries = idageodf_county['SHAPE'].centroid()
        assert(isinstance(idageoseries.y(), IdaSeries))

    def test_idageoseries_z(self, idageodf_county):
        idageoseries = idageodf_county['SHAPE'].centroid()
        assert(isinstance(idageoseries.z(), IdaSeries))

    def test_idageoseries_is_closed(self, idageodf_county):
        # TODO: add dataset with a column with ST_LINESTRING
        pass

    def test_idageoseries_is_empty(self, idageodf_county):
        idageoseries = idageodf_county['SHAPE']
        assert(isinstance(idageoseries.is_empty(), IdaSeries))

    def test_idageoseries_is_simple(self, idageodf_county):
        idageoseries = idageodf_county['SHAPE']
        assert(isinstance(idageoseries.is_simple(), IdaSeries))
    
    def test_idageoseries_check_linear_unit(self, idageodf_county):
        idageoseries = idageodf_county['SHAPE']
        with pytest.raises(TypeError):
            idageoseries._check_linear_unit(10)
        with pytest.raises(IdaGeoDataFrameError):
            idageoseries._check_linear_unit('not a valid unit')
        unit = 'meter'
        ans = idageoseries._check_linear_unit(unit)
        assert(ans == '\'METER\'')
        unit = 'Yard (SEARS)' # parenthesis
        ans = idageoseries._check_linear_unit(unit)
        assert(ans == '\'YARD (SEARS)\'')
        unit = 'Clarke\'s foot' # quotation mark
        ans = idageoseries._check_linear_unit(unit)
        assert(ans == '\'CLARKE\'\'S FOOT\'')
        unit = 'bin width 37.5 METRES' #with dot    
        ans = idageoseries._check_linear_unit(unit)
        assert(ans == '\'BIN WIDTH 37.5 METRES\'')
    
    def test_idageoseries_linear_units(self, idageodf_county):
        idageoseries = idageodf_county['SHAPE']
        units = idageoseries.linear_units
        assert(not units.empty)

    def test_idageoseries_unary_operation_handler_non_geometry_column(
            self, idageodf_county):
        idageoseries = idageodf_county['SHAPE'] # ST_POLYGON      
        with pytest.raises(TypeError):
            idageoseries._unary_operation_handler(
                db2gse_function = 'DB2GSE.ST_AGEOSPATIALFUNCTION',
                valid_types = ['ST_POINT'])

    def test_idageoseries_area_km(self, idageodf_county):
        idageodf = idageodf_county
        idageodf.set_geometry('SHAPE')
        idageodf['AREA_KM'] = idageodf.area(unit='KILOMETER')
        max_area_km = idageodf['AREA_KM'].max()
        assert(int(max_area_km)== 52073)

    def test_idageoseries_area_km_lc(self, idageodf_county_view):
        idageodf = idageodf_county_view
        idageodf.set_geometry('shape')
        idageodf['area_km'] = idageodf.area(unit='kilometer')
        max_area_km = idageodf['area_km'].max()
        assert(int(max_area_km)== 52073)

    def test_idageoseries_area_km_mbr(self, idageodf_county):
        idageodf = idageodf_county
        idageodf.set_geometry('SHAPE')
        idageodf['MBR'] = idageodf.mbr()
        idageodf.set_geometry('MBR')
        idageodf['AREA_KM_MBR'] = idageodf.area(unit='KILOMETER')
        max_area_km_mbr = idageodf['AREA_KM_MBR'].max()
        assert(int(max_area_km_mbr) == 100246)

    def test_idageoseries_area_km_mbr_lc(self, idageodf_county_view):
        idageodf = idageodf_county_view
        idageodf.set_geometry('shape')
        idageodf['mbr'] = idageodf.mbr()
        idageodf.set_geometry('mbr')
        idageodf['area_km_mbr'] = idageodf.area(unit='kilometer')
        max_area_km_mbr = idageodf['area_km_mbr'].max()
        assert(int(max_area_km_mbr) == 100246)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright (c) 2015, IBM Corp.
# All rights reserved.
#
# Distributed under the terms of the BSD Simplified License.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from future import standard_library
standard_library.install_aliases()
from .base import IdaDataBase
from .frame import IdaDataFrame
from .series import IdaSeries
from .geoFrame import IdaGeoDataFrame
from .geoSeries import IdaGeoSeries

__all__ = ['learn', 'sampledata', 'tests', 'aggregation', 
		   'base', 'exceptions', 'filtering', 'frame', 'indexing', 
		   'internals', 'series', 'sql', 'statistics', 'utils', 'geoFrame',
             'geoSeries']



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright (c) 2015, IBM Corp.
# All rights reserved.
#
# Distributed under the terms of the BSD Simplified License.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------

# For local installation do "pip install -e" in the ibmdbpy-master directory
###############################################################################

"""
An IdaDataBase instance represents a reference to a remote Netezza Warehouse database
maintaining attributes and methods for administration of the database.
"""

from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division
from __future__ import absolute_import
from builtins import zip
from builtins import int
from builtins import str
from future import standard_library
standard_library.install_aliases()

import os
#from os import path
import sys
import math
import random
from time import time
import datetime
import warnings
from copy import deepcopy

from collections import OrderedDict


import numpy as np
import pandas as pd
from pandas.io.sql import read_sql

from lazy import lazy
import six

import nzpyida
from nzpyida import sql
from nzpyida.utils import timed, set_verbose, set_autocommit
from nzpyida.exceptions import IdaDataBaseError, PrimaryKeyError

class IdaDataBase(object):
    """
    An IdaDataBase instance represents a reference to a remote Netezza Warehouse
    database. This is an abstraction layer for the remote connection. The
    IdaDataBase interface provides several functions that enable basic database 
    administration in pythonic syntax.

    You can use either ODBC or JDBC or NZPY to connect to the database. The default
    connection type is ODBC, which is the standard connection type for Windows 
    users. To establish an ODBC connection, download an IBM Netezza driver and set
    up your ODBC connection by specifying your connection protocol, port, and 
    hostname. An ODBC connection on Linux or Mac might require more settings. 
    For more information about how to establish an ODBC connection, see the 
    pypyodbc documentation.

    To connect with JDBC, install the optional external package jaydebeapi, 
    download the ibm jdbc driver, and save it in your local nzpyida folder.
    If you put the jdbc driver in the CLASSPATH variable or the folder that
    contains it, it will work too. A C++ compiler adapted to the current python 
    version, operating system, and architecture may also be required to install
    jaydebeapi.

    To connect with NZPY, specify the required parameters in dict format, there is
    no need to install any drivers.

    The instantiation of an IdaDataBase object is a mandatory step before 
    creating IdaDataFrame objects because IdaDataFrames require an IdaDataBase 
    as a parameter to be initialized. By convention, use only one instance of 
    IdaDataBase per database. However, you can use several instances of 
    IdaDataFrame per connection.
    """

    def __init__(self, dsn, uid='', pwd='', autocommit=True, verbose=False):
        """
        Open a database connection.

        Parameters
        ----------
        dsn : str
            Data Source Name (as specified in your ODBC settings) or JDBC URL 
            string or NZPY dict

        uid : str, optional
            User ID.

        pwd : str, optional
            User password.

        autocommit : bool, default: True
            If True, automatically commits all operations.

        verbose : bool, defaukt: True
            If True, prints all SQL requests that are sent to the database. 

        Attributes
        ----------
        data_source_name : str
            Name of the referring DataBase.

        _con_type : str
            Type of the connection, either 'odbc' or 'jdbc' or 'nzpy'.

        _connection_string : str
            Connection string use for connecting via ODBC or JDBC or NZPY.

        _con : connection object
            Connection object to the remote Database.

        _database_system: str
            Underlying database system, either 'db2' or 'netezza'

        _database_name: str
            The name of the database the application is connected to.

        _idadfs : list
            List of IdaDataFrame objects opened under this connection.

        Returns
        -------
        IdaDataBase object

        Raises
        ------
        ImportError
            JayDeBeApi is not installed.
        IdaDataBaseError
            * uid and pwd are defined both in uid, pwd parameters and dsn.
            * The 'db2jcc4.jar' file is not in the ibmdbpy site-package repository.

        Examples
        --------

        ODBC connection, userID and password are stored in ODBC settings:

        >>> IdaDataBase(dsn="BLUDB") # ODBC Connection
        <ibmdbpy.base.IdaDataBase at 0x9bec860>

        ODBC connection, userID and password are not stored in ODBC settings:

        >>> IdaDataBase(dsn="BLUDB", uid="<UID>", pwd="<PWD>")
        <ibmdbpy.base.IdaDataBase at 0x9bec860>
        
        JDBC connection, full JDBC string:

        >>> jdbc='jdbc:db2://<HOST>:<PORT>/<DBNAME>:user=<UID>;password=<PWD>'
        >>> IdaDataBase(dsn=jdbc)
        <ibmdbpy.base.IdaDataBase at 0x9bec860>

        JDBC connectiom, JDBC string and seperate userID and password:

        >>> jdbc = 'jdbc:db2://<HOST>:<PORT>/<DBNAME>'
        >>> IdaDataBase(dsn=jdbc, uid="<UID>", pwd="<PWD>")
        <ibmdbpy.base.IdaDataBase at 0x9bec860>


         NZPY connection, userID and password are stored in NZPY dict:

         nzpy_dsn ={
        "database":"<DBNAME>",
         "port" :<PORT>,
        "host" : "<HOST>",
        "securityLevel":<SECURITYLEVEL>,
        "logLevel":<LOGLEVEL>,
        "user":"<UID>",
        "password":"<PWD>"
        }

        >>> IdaDataBase(dsn=nzpy_dsn)
        <ibmdbpy.base.IdaDataBase at 0x9bec860>


        NZPY connection, NZPY dict and seperate userID and password

         nzpy_dsn ={
        "database":"<DBNAME>",
         "port" :<PORT>,
        "host" : "<HOST>",
        "securityLevel":<SECURITYLEVEL>,
        "logLevel":<LOGLEVEL>

        }

        >>> IdaDataBase(dsn=nzpy_dsn, uid="<UID>", pwd="<PWD>")



        """

        if isinstance(dsn, dict)==False:
            for arg,name in zip([dsn, uid, pwd],['dsn','uid','pwd']):
              if not isinstance(arg, six.string_types):
                raise TypeError("Argument '%s' of type %, expected : string type."%(name,type(arg)))

        self.data_source_name = dsn


        # default value for _database_system is db2
        self._database_system = 'db2'
        # first delimiter before the parameters in the jdbc-url
        url_1stparam_del = ':'


        # Detect if user attempt to connection with ODBC or JDBC
        if isinstance(dsn,str):
            if dsn.startswith('jdbc:'):
              self._con_type = "jdbc"
              if dsn.startswith('jdbc:netezza:'):
                  self._database_system = 'netezza'
                  # for Netezza the internal connection is always in autocommit mode
                  # to allow explicit commits
                  dsn += ';autocommit=false'
                  url_1stparam_del = ';'
              elif dsn.startswith('jdbc:db2:'):
                  self._database_system = 'db2'
              else:
                  raise IdaDataBaseError(("The JDBC connection string is invalid for Db2 and Netezza. " +
                                          "It has to start either with 'jdbc:db2:' or 'jdbc:netezza:'."))

            else:
                self._con_type='odbc'

        elif isinstance(dsn, dict):
            self._con_type = "nzpy"


        self._idadfs = []



        if self._con_type == 'nzpy':
            #parse dict
            host=''
            port=0
            database=''
            securityLevel=0
            logLevel=0
            user =''
            password=''
            for key,value in dsn.items():
                if(key=='host'):
                    host=value
                if (key=='port'):
                    port=value
                if (key=='database'):
                    database=value
                if (key=='securityLevel') :
                    securityLevel=value
                if (key=='logLevel') :
                    logLevel=value
                if (key == 'user'):
                    user = value
                if (key == 'password'):
                    password = value

            import nzpy;
            missingCredentialsMsg = ("Missing credentials to connect via NZPY.")
            ambiguousDefinitionMsg = ("Ambiguous definition of userID or password: " +
                                      "Cannot be defined in uid and pwd parameters " +
                                      "and in nzpy dict at the same time.")


            uidSpecified = uid != ''
            pwdSpecified = pwd != ''

            if not ('user' in dsn):
                if (uidSpecified):
                    user= uid
            elif (uidSpecified):
                raise IdaDataBaseError(ambiguousDefinitionMsg)
            uidSpecified = uidSpecified | ('user' in dsn)

            if not ('password' in dsn):
                if (pwdSpecified):
                   password= pwd
            elif (pwdSpecified):
                raise IdaDataBaseError(ambiguousDefinitionMsg)
            pwdSpecified = pwdSpecified | ('password' in dsn)

            # throw an exception if either uid or pwd are specified,
            # i.e. not both or none of the two
            if (uidSpecified ^ pwdSpecified):
                raise IdaDataBaseError(missingCredentialsMsg)



            try:
                if port <= 0:
                    port = 5480
                self._con = nzpy.connect(user=user, password=password,host=host, port=port,
                                        database=database, securityLevel=securityLevel,logLevel=logLevel)
            except Exception as e:
                raise IdaDataBaseError(str(e))
            self._connection_string ={'user':user,'password':password,'host':host,
                'port':port, 'database':database, 'securityLevel':securityLevel,'logLevel':logLevel}
            self._database_system = 'netezza'

        if self._con_type == 'odbc' :

            self._connection_string = "DSN=%s; UID=%s; PWD=%s;LONGDATACOMPAT=1;"%(dsn,uid,pwd)
            """
            Workaround for CLOB retrieval: 
            Set the CLI/ODBC LongDataCompat keyword to 1. 
            Doing so will force the CLI driver to make the 
            following data type mappings
            SQL_CLOB to SQL_LONGVARCHAR
            SQL_BLOB to SQL_LONGVARBINARY
            SQL_DBCLOB to SQL_WLONGVARCHAR
            """

            import pyodbc
            try:
                self._con = pyodbc.connect(self._connection_string)
                self._con.setdecoding(pyodbc.SQL_CHAR, encoding='utf-8')
                self._con.setdecoding(pyodbc.SQL_WCHAR, encoding='utf-8')
                self._con.setencoding(encoding='utf-8')
            except Exception as e:
                raise IdaDataBaseError(str(e))
            try:
                self.ida_query("select count(*) from _V_OBJECT")
                self._database_system = 'netezza'
            except  Exception as e1:
                try:
                    self.ida_query("select CURRENT_SERVER from SYSIBM.SYSDUMMY1")
                    self._database_system = 'db2'
                except Exception as e2:
                    errorMsg = ("The following errors occurred when trying to determine " +
                                "if the database system is Netezza or Db2:\n" +
                                "%s\n%s") % (str(e1), str(e2))
                    raise IdaDataBaseError(errorMsg)

        if self._con_type == 'jdbc':

            missingCredentialsMsg = ("Missing credentials to connect via JDBC.")
            ambiguousDefinitionMsg = ("Ambiguous definition of userID or password: " +
                                      "Cannot be defined in uid and pwd parameters " +
                                      "and in jdbc_url_string at the same time.")

            # remove trailing ":" or ";" or spaces on dsn, we will replace.
            dsn = dsn.rstrip(';: ')
            # find parameters on dsn; if any exist, there will be an equals sign.
            ix = dsn.find("=")

            # if no parameters exist, then this is the complete dsn
            if (ix < 0):
                # nothing needs to be done, if uid and pwd are missing
                # (uid and pwd are not needed for local database connections)
                # otherwise both uid and pwd have to be specified
                if uid and pwd:
                    dsn = dsn + url_1stparam_del + 'user={};password={};'.format(uid, pwd)
                elif not uid and not pwd:
                   # Neither UID nor PWD have to be specified for local connections.
                   # This assumes that if they're missing, the connection is local.
                    pass
                else:
                    raise IdaDataBaseError(missingCredentialsMsg)
            else:
                # if we know there is at least one parameter; we can assume there exists a ":" before the parameter
                # portion of the string in a correctly formatted dsn.  Therefore, just check for the existence of 
                # the uid and pwd and add if they are missing.  If they are on the string, IGNORE the string as-is.

                uidSpecified = uid != ''
                pwdSpecified = pwd != ''

                if not ('user=' in dsn):
                    if (uidSpecified):
                       dsn = dsn + ';user=' + uid
                elif (uidSpecified):
                    raise IdaDataBaseError(ambiguousDefinitionMsg)
                uidSpecified = uidSpecified | ('user=' in dsn)

                if not ('password=' in dsn):
                    if (pwdSpecified):
                        dsn = dsn + ';password=' + pwd
                elif (pwdSpecified):
                    raise IdaDataBaseError(ambiguousDefinitionMsg)
                pwdSpecified = pwdSpecified | ('password=' in dsn)

                # throw an exception if either uid or pwd are specified,
                # i.e. not both or none of the two
                if (uidSpecified^pwdSpecified):
                    raise IdaDataBaseError(missingCredentialsMsg)

                 # add trailing ";" to dsn.
                dsn = dsn + ';'

            jdbc_url = dsn

            try:
                import jaydebeapi
                import jpype
            except ImportError:
                raise ImportError("Please install optional dependency jaydebeapi "+
                            "to work with JDBC.")

            # check versions
            if jaydebeapi.__version__.startswith('0'):
                # Older JaydeBeAPi versions where the connection information is specified through a list
                # are not supported anymore.
                # The connection information has to be included now in a single connection string.
                message = ("Your JayDeBeApi module is not supported anymore.  Please install version 1.x or higher.")
                raise IdaDataBaseError(message)

            if jpype.__version__ != '0.6.3':
                message = ("Your JPype1 version is not compatible with JayDeBeApi. Please install version 0.6.3.")
                raise IdaDataBaseError(message)

            here = os.path.abspath(os.path.dirname(__file__))
            driver_not_found = ""
            if self._is_netezza_system():
                driverlibs = ["nzjdbc3.jar"]
            else:
                driverlibs = ["db2jcc4.jar", "db2jcc.jar"]

            if (not jpype.isJVMStarted()):
                classpath = os.getenv('CLASSPATH','')
                jarpath = ''
                platform = sys.platform

                if verbose: print("Trying to find a path to the JDBC driver jar file in CLASSPATH (%s)."%classpath)

                if platform == 'win32':
                    classpaths = classpath.split(';')
                else:
                    classpaths = classpath.split(':')

                if any(dl for dl in driverlibs if dl in classpath):
                    # A path to the driver exists in the classpath variable
                    jarpaths =  [jp for jp in classpaths if any([dl in jp for dl in driverlibs])]
                    jarpath = ""
                    
                    while jarpaths: # check if a least one path exists
                        jarpath = jarpaths.pop(0) # just take the first
                        if os.path.isfile(jarpath): # Is the path correct?
                            if platform == 'win32':
                                jarpath = jarpath.split(':')[1].replace('\\', '/') # get rid of windows style formatting
                            break
                        else:
                            if verbose: print("The path %s does not seem to be correct.\nTrying to recover..."%jarpath)
                            jarpath = ""

                if not jarpath: # Means the search for a direct path in the CLASSPATH was not successful
                    def _get_jdbc_driver_from_folders(folders):
                        jarpath = ''
                        if platform == 'win32':
                            jarpaths = [fld.split(':')[1].replace('\\', '/') + "/" + dl
                                        for fld in folders for dl in driverlibs if os.path.isfile(fld + "/" + dl)]
                            if jarpaths == []:
                                # check if classpath contains paths with wildcard "*" at the end
                                jarpaths = [fld.split(':')[1].replace('\\', '/')[:-1] + dl
                                            for fld in folders for dl in driverlibs if
                                            fld[-1:] == '*' and os.path.isfile(fld + "/" + dl)]
                        else:
                            jarpaths = [fld + "/" + dl for fld in folders for dl in driverlibs if
                                        os.path.isfile(fld + "/" + dl)]
                            if jarpaths == []:
                                # check if classpath contains paths with wildcard "*" at the end
                                jarpaths = [fld[:-1] + dl
                                            for fld in folders for dl in driverlibs if
                                            fld[-1:] == '*' and os.path.isfile(fld[:-1] + dl)]
                        if jarpaths:
                            return jarpaths[0]
                        
                    if classpath: # There is at least something in the classpath variable
                        # Let us see if the jar in a folder of the classpath
                        if verbose: print("Trying to find the JDBC driver in the folders of CLASSPATH (%s)."%classpath)

                        jarpath = _get_jdbc_driver_from_folders(classpaths)
                    
                        if not jarpath: # jarpath is still ''
                            if verbose: print("Trying to find the JDBC driver in the local folder of ibmdbpy (%s)"%here)
                            # Try to get the path to the driver from the ibmdbpy folder, the last chance
                            jarpath = _get_jdbc_driver_from_folders([here])

                if jarpath:
                    if verbose: print("Found it at %s!\nTrying to connect..."%jarpath)

                jpype.startJVM(jpype.getDefaultJVMPath(), '-Djava.class.path=%s' % jarpath)


            if self._is_netezza_system():
                driver_not_found = ("HELP: The Netezza JDBC driver library 'nzjdbc3.jar' could not be found. "+
                                    "Please, follow the instructions on "+
                                    "'https://www.ibm.com/support/knowledgecenter/en/SS5FPD_1.0.0/com.ibm.ips.doc/postgresql/odbc/c_datacon_plg_overview.html' "+
                                    "for downloading and installing the JDBC driver "+
                                    "and put the file 'nzjdbc3.jar' in the CLASSPATH variable "+
                                    "or in a folder that is in the CLASSPATH variable. Alternatively place "+
                                    "it in the folder '%s'."%here)

            else:
                driver_not_found = ("HELP: The JDBC driver for IBM Db2 could "+
                                    "not be found. Please download the latest JDBC Driver at the "+
                                    "following address: 'https://www.ibm.com/support/pages/node/382667' "+
                                    "and put the file 'db2jcc.jar' or 'db2jcc4.jar' in the CLASSPATH variable "+
                                    "or in a folder that is in the CLASSPATH variable. Alternatively place "+
                                    "it in the folder '%s'."%here)

            self._connection_string = jdbc_url

            try:
                if self._is_netezza_system():
                    self._con = jaydebeapi.connect('org.netezza.Driver', self._connection_string)
                else:
                    self._con = jaydebeapi.connect('com.ibm.db2.jcc.DB2Driver', self._connection_string)
            except  Exception as e:
                print(driver_not_found)
                raise IdaDataBaseError(e)
             
            if verbose: print("Connection successful!")
                
            #add DB2GSE to the database FUNCTION PATH            
            #query = "SET CURRENT FUNCTION PATH = CURRENT FUNCTION PATH, db2gse"
            #self.ida_query(query)
            #not anymore, reported problems with ODBC
            #better mention DB2GSE explicitly when accessing its functions

        # determine name of database
        if self._is_netezza_system():
            self._database_name = self.ida_scalar_query('select OBJNAME from _T_OBJECT where OBJID = CURRENT_DB;')
        else:
            self._database_name = self.ida_scalar_query('select CURRENT_SERVER from SYSIBM.SYSDUMMY1')

        # Setting Autocommit and verbose environment variables
        set_autocommit(autocommit)
        set_verbose(verbose)

    ###########################################################################
    #### Data Exploration
    ###########################################################################

    @lazy
    def current_schema(self):
        """
        Get the current user schema name as a string.

        Returns
        -------
        str
            User's schema name.

        Examples
        --------
        >>> idadb.current_schema()
        'DASHXXXXXX'
        """
        if self._is_netezza_system():
          query = 'select TRIM(CURRENT_SCHEMA)'
        else:
          query = "SELECT TRIM(CURRENT_SCHEMA) FROM SYSIBM.SYSDUMMY1"
        return self.ida_scalar_query(query)

    def show_tables(self, show_all=False):
        """
        Show tables and views that are available in self. By default, this 
        function shows only tables that belong to a user’s specific schema.

        Parameters
        ----------
        show_all : bool
            If True, all table and view names in the database are returned, 
            not only those that belong to the user's schema.

        Returns
        -------
        DataFrame
             A data frame containing tables and views names in self with some 
             additional information (TABSCHEMA, TABNAME, OWNER, TYPE). 

        Examples
        --------
        >>> ida_db.show_tables()
             TABSCHEMA           TABNAME       OWNER TYPE
        0    DASHXXXXXX            SWISS  DASHXXXXXX    T
        1    DASHXXXXXX             IRIS  DASHXXXXXX    T
        2    DASHXXXXXX     VIEW_TITANIC  DASHXXXXXX    V
        ...
        >>> ida_db.show_tables(show_all = True)
             TABSCHEMA           TABNAME       OWNER TYPE
        0    DASHXXXXXX            SWISS  DASHXXXXXX    T
        1    DASHXXXXXX             IRIS  DASHXXXXXX    T
        2    DASHXXXXXX     VIEW_TITANIC  DASHXXXXXX    V
        2      SYSTOOLS      IDAX_MODELS  DASH101631    A
        ...

        Notes
        -----
        show_tables implements a cache strategy. The cache is stored when the 
        user calls the method with the argument show_all set to True. This 
        improves performance because database table look ups are a very common 
        operation. The cache gets updated each time a table or view is created 
        or refreshed, each time a table or view is deleted, or when a new 
        IdaDataFrame is opened.

        """
        #### DEVELOPERS FIX: UNCOMMENT WHEN DEAD LOCK
        #show_all = False
        ####

        # Try to retrieve the cache
        if show_all:
            cache = self._retrieve_cache("cache_show_tables")
            if cache is not None:
                return cache

        where_part = ""
        if not show_all:
            where_part = ("AND TABSCHEMA = '%s' " % self.current_schema)

        if self._is_netezza_system():
            query = ("SELECT SCHEMA as TABSCHEMA, TABLENAME as TABNAME, OWNER, 'T' as TYPE FROM _V_TABLE " +
                     "WHERE DATABASE = '" + self._database_name + "' and OBJTYPE = 'TABLE' " + where_part +
                     "UNION ALL " +
                     "SELECT SCHEMA as TABSCHEMA, VIEWNAME as TABNAME, OWNER, 'V' as TYPE FROM _V_VIEW " +
                     "WHERE DATABASE = '" + self._database_name + "' and OBJTYPE = 'VIEW' " + where_part)
            query = ("SELECT SCHEMA as TABSCHEMA, OBJNAME as TABNAME, OWNER, " +
                             "CASE WHEN OBJTYPE = 'TABLE' THEN 'T' ELSE 'V' END AS TYPE " +
                      "FROM _V_OBJECTS  " +
                      "WHERE OBJTYPE in ('TABLE', 'VIEW') " + where_part +
                      "ORDER BY TABSCHEMA, TABNAME")
        else:
            query = ("SELECT DISTINCT TABSCHEMA, TABNAME, OWNER, TYPE " +
                     "FROM SYSCAT.TABLES " +
                     "WHERE OWNERTYPE = 'U' " + where_part +
                     "ORDER BY TABSCHEMA, TABNAME")

        data = self.ida_query(query)
        
        # Workaround for some ODBC version which does not get the entire
        # string of the column name in the cursor descriptor. 
        # This is hardcoded, so be careful
        data.columns = ['TABSCHEMA', 'TABNAME', 'OWNER', 'TYPE']
        
        data = self._upper_columns(data)

        # Db2 Warehouse FIX: schema "SAMPLES" and "GOSALES" saved with an extra blank,
        # By doing so, we delete the extra blank.
        # Note that this works because all cells are of type string.

        # OLD version, created an unexpected bug in some wrong ODBC version
        # TypeError: unorderable types: bytes() > int()
        # less pythonic, do not use anymore 
        #for col in data:
        #    for index, val in enumerate(data[col]):
        #        data[col][index] = val.strip()

        # Can be done with a one liner :) 
        data = data.apply(lambda x: x.apply(lambda x: x.strip()))

        # Cache the result
        if show_all is True:
            self.cache_show_tables = data

        return data

    def show_models(self):
        """
        Show models that are available in the database.

        Returns
        -------
        DataFrame

        Examples
        --------
        >>> idadb.show_models()
            MODELSCHEMA               MODELNAME       OWNER
        0   DASHXXXXXX  KMEANS_10857_1434974511  DASHXXXXXX
        1   DASHXXXXXX  KMEANS_11726_1434977692  DASHXXXXXX
        2   DASHXXXXXX  KMEANS_11948_1434976568  DASHXXXXXX
        """
        if self._is_netezza_system():
            list_models_stmt = ("SELECT MODELNAME, OWNER, CREATED, STATE, MININGFUNCTION, ALGORITHM, USERCATEGORY " +
                                "FROM INZA.V_NZA_MODELS")
            result_columns =  ['modelname', 'owner', 'created', 'state','miningfunction', 'algorithm', 'usercategory']
        else:
            list_models_stmt = "call IDAX.LIST_MODELS()"
            result_columns = ['modelschema', 'modelname', 'owner', 'created', 'state',
                              'miningfunction', 'algorithm', 'usercategory']

        data = self.ida_query(list_models_stmt)

        data.columns = result_columns

        # Workaround for some ODBC version which does not get the entire
        # string of the column name in the cursor descriptor. 
        # This is hardcoded, so be careful
        data = self._upper_columns(data)
        return data

    def exists_table_or_view(self, objectname):
        """
        Check if a table or view exists in self.

        Parameters
        ----------
        objectname : str
            Name of the table or view to check.

        Returns
        -------
        bool

        Raises
        ------
        TypeError
            The object exists but is not of the expected type.

        Examples
        --------
        >>> idadb.exists_table_or_view("NOT_EXISTING")
        False
        >>> idadb.exists_table_or_view("TABLE_OR_VIEW")
        True
        >>> idadb.exists_table_or_view("NO_TABLE_NOR_VIEW")
        TypeError : "NO_TABLE_NOR_VIEW" exists in schema '?' but of type '?'
        """
        return self._exists(objectname,['T', 'V'])

    def exists_table(self, tablename):
        """
        Check if a table exists in self.

        Parameters
        ----------
        tablename : str
            Name of the table to check.

        Returns
        -------
        bool

        Raises
        ------
        TypeError
            The object exists but is not of the expected type.

        Examples
        --------
        >>> idadb.exists_table("NOT_EXISTING")
        False
        >>> idadb.exists_table("TABLE")
        True
        >>> idadb.exists_table("NO_TABLE")
        TypeError : "tablename" exists in schema "?" but of type '?'
        """
        return self._exists(tablename,['T'])

    def exists_view(self, viewname):
        """
        Check if a view exists in self.

        Parameters
        ----------
        viewname : str
            Name of the view to check.

        Returns
        -------
        bool

        Raises
        ------
        TypeError
            The object exists but is not of the expected type.

        Examples
        --------
        >>> idadb.exists_view("NOT_EXISTING")
        False
        >>> idadb.exists_view("VIEW")
        True
        >>> idadb.exists_view("NO_VIEW")
        TypeError : "viewname" exists in schema "?" but of type '?'
        """
        return self._exists(viewname,['V'])

    def exists_model(self, modelname):
        """
        Check if a model exists in self.

        Parameters
        ----------
        modelname : str
            Name of the model to check. It should contain only alphanumeric
            characters and underscores. All lower case characters will be
            converted to upper case characters.

        Returns
        -------
        bool

        Raises
        ------
        TypeError
            The object exists but is not of the expected type.

        Examples
        --------
        >>> idadb.exists_model("MODEL")
        True
        >>> idadb.exists_model("NOT_EXISTING")
        False
        >>> idadb.exists_model("NO_MODEL")
        TypeError : NO_MODEL exists but is not a model (of type '?')
        """
        modelname = nzpyida.utils.check_modelname(modelname)
        if '.' in modelname:
            modelschema, modelname = modelname.split('.')
        else:
            modelschema = self.current_schema

        if self._is_netezza_system():
            # on Netezza the model schema part of the model name is ignored
            modelquery = "SELECT count(*) FROM INZA.V_NZA_MODELS WHERE MODELNAME ='%s'"%modelname
            modelexists = self.ida_scalar_query(modelquery) >= 1
            if modelexists:
                return True
        else:
            # check if schema exists to avoid exception thrown by idax.list_models
            schemaquery = "SELECT count(*) FROM SYSCAT.SCHEMATA WHERE SCHEMANAME = '%s'"%modelschema
            schemaexists = self.ida_scalar_query(schemaquery) >= 1
            if not schemaexists:
                return False
            data = self.ida_query("call idax.list_models('schema=%s, where=MODELNAME=''%s''')" % (modelschema, modelname))
            if not data.empty:
                return True

        tablelist = self.show_tables(show_all=True)
        tablelist = tablelist[(tablelist['TABSCHEMA']==modelschema) & (tablelist['TABNAME']==modelname)]
        if len(tablelist):
             tabletype = tablelist['TYPE'].values[0]
             raise TypeError("%s.%s exists, but is not a model (of type '%s')"
                             % (modelschema, modelname, tabletype))

        else:
            return False

    def is_table_or_view(self, objectname):
        """
        Check if an object is a table or a view in self.

        Parameters
        ----------
        objectname : str
            Name of the object to check.

        Returns
        -------
        bool

        Raises
        ------
        ValueError
            objectname doesn't exist in the database.


        Examples
        --------
        >>> idadb.is_table_or_view("NO_TABLE")
        False
        >>> idadb.is_table_or_view("TABLE")
        True
        >>> idadb.is_table_or_view("NOT_EXISTING")
        ValueError : NO_EXISTING does not exist in database
        """
        return self._is(objectname,['T','V'])

    def is_table(self, tablename):
        """
        Check if an object is a table in self.

        Parameters
        ----------
        tablename : str
            Name of the table to check.

        Returns
        -------
        bool

        Raises
        ------
        ValueError
            The object doesn't exist in the database.


        Examples
        --------
        >>> idadb.is_table("NO_TABLE")
        False
        >>> idadb.is_table("TABLE")
        True
        >>> idadb.is_table("NOT_EXISTING")
        ValueError : NO_EXISTING does not exist in database
        """
        return self._is(tablename,['T'])

    def is_view(self, viewname):
        """
        Check if an object is a view in self.

        Parameters
        ----------
        viewname : str
            Name of the view to check.

        Returns
        -------
        bool

        Raises
        ------
        ValueError
            The object doesn't exist in the database.

        Examples
        --------
        >>> idadb.is_view("NO_VIEW")
        False
        >>> idadb.is_view("VIEW")
        True
        >>> idadb.is_view("NOT_EXISTING")
        ValueError : NO_EXISTING does not exist in database
        """
        return self._is(viewname,['V'])

    def is_model(self, modelname):
        """
        Check if an object is a model in self.

        Parameters
        ----------
        modelname : str
            Name of the model to check. It should contain only alphanumeric
            characters and underscores. All lower case characters will be
            converted to upper case characters.

        Returns
        -------
        bool

        Raises
        ------
        ValueError
            The object doesn't exist in the database.

        Examples
        --------
        >>> idadb.is_model("MODEL")
        True
        >>> idadb.is_model("NO_MODEL")
        False
        >>> idadb.is_model("NOT_EXISTING")
        ValueError : NOT_EXISTING doesn't exist in database
        """
        modelname = nzpyida.utils.check_modelname(modelname)

        if '.' in modelname:
            modelname_noschema = modelname.split('.')[-1]
        else:
            modelname_noschema = modelname
        data = self.show_models()
        if not data.empty:
            if modelname_noschema in data['MODELNAME'].values:
                return True

        # This part is executed if data is empty or model is not in data
        try:
            self.is_table_or_view(modelname)
        except:
            raise
        else:
            return False

    def ida_query(self, query, silent=False, first_row_only=False, autocommit = False):
        """
        Prepare, execute and format the result of a query in a dataframe or
        in a Tuple. If nothing is expected to be returned for the SQL command,
        nothing is returned. 

        Parameters
        ----------
        query : str
            Query to be executed.
        silent: bool, default: False
             If True, the query is not printed in the python console even if 
             the verbosity mode is activated (VERBOSE environment variable is 
             equal to “True”). 
        first_row_only : bool, default: False
             If True, only the first row of the result is returned as a Tuple.
        autocommit: bool, default: False
            If True, the autocommit function is available.

        Returns
        -------
        DataFrame or Tuple (if first_row_only=False)

        Examples
        --------
        >>> idadb.ida_query("SELECT * FROM IRIS LIMIT 5")
           sepal_length  sepal_width  petal_length  petal_width species
        0           5.1          3.5           1.4          0.2  setosa
        1           4.9          3.0           1.4          0.2  setosa
        2           4.7          3.2           1.3          0.2  setosa
        3           4.6          3.1           1.5          0.2  setosa
        4           5.0          3.6           1.4          0.2  setosa

        >>> idadb.ida_query("SELECT COUNT(*) FROM IRIS")
        (150, 150, 150, 150)

        Notes
        -----
        If first_row_only argument is True, then even if the actual result of 
        the query is composed of several rows, only the first row will be 
        returned.
        """
        self._check_connection()
        return sql.ida_query(self, query, silent, first_row_only, autocommit)

    def ida_scalar_query(self, query, silent=False, autocommit = False):
        """
        Prepare and execute a query and return only the first element as a 
        string. If nothing is returned from the SQL query, an error occurs.

        Parameters
        ----------
        query : str
            Query to be executed.
        silent: bool, default: False
            If True, the query will not be printed in python console even if
            verbosity mode is activated.
        autocommit: bool, default: False
            If True, the autocommit function is available.

        Returns
        -------
        str or Number

        Examples
        --------
        >>> idadb.ida_scalar_query("SELECT TRIM(CURRENT_SCHEMA) from SYSIBM.SYSDUMMY1")
        'DASHXXXXX'

        Notes
        -----
        Even if the actual result of the query is composed of several columns
        and several rows, only the first element (top-left) will be returned.
        """
        self._check_connection()
        return sql.ida_scalar_query(self, query, silent, autocommit)

    ###############################################################################
    #### Upload DataFrames
    ###############################################################################

    @timed
    def as_idadataframe(self, dataframe, tablename=None, clear_existing=False, primary_key=None, indexer=None):
        """
        Upload a dataframe and return its corresponding IdaDataFrame. The target
        table (tablename) will be created or replaced if the option clear_existing
        is set to True.
        
        To add data to an existing tables, see IdaDataBase.append

        Parameters
        ----------
        dataframe : DataFrame
            Data to be uploaded, contained in a Pandas DataFrame.
        tablename : str, optional
            Name to be given to the table created in the database.
            It should contain only alphanumeric characters and underscores.
            All lower case characters will be converted to upper case characters.
            If not given, a valid tablename is generated (for example, DATA_FRAME_X
            where X is a random number).
        clear_existing : bool
            If set to True, a table will be replaced when a table with the same 
            name already exists  in the database.
        primary_key : str
            Name of a column to be used as primary key.

        Returns
        -------
        IdaDataFrame

        Raises
        ------
        TypeError
            * Argument dataframe is not of type pandas.DataFrame.
            * The primary key argument is not a string.

        NameError
            * The name already exists in the database and clear_existing is False.
            * The primary key argument doesn't correspond to a column.

        PrimaryKeyError
            The primary key contains non unique values.

        Examples
        --------
        >>> from nzpyida.sampledata.iris import iris
        >>> idadb.as_idadataframe(iris, "IRIS")
        <ibmdbpy.frame.IdaDataFrame at 0xb34a898>
        >>> idadb.as_idadataframe(iris, "IRIS")
        NameError: IRIS already exists, choose a different name or use clear_existing option.
        >>> idadb.as_idadataframe(iris, "IRIS2")
        <ibmdbpy.frame.IdaDataFrame at 0xb375940>
        >>> idadb.as_idadataframe(iris, "IRIS", clear_existing = True)
        <ibmdbpy.frame.IdaDataFrame at 0xb371cf8>
        
        Notes
        -----
        This function is not intended to be used to add data to an existing table,
        rather to create a new table from a dataframe. To add data to an existing
        table, please consider using IdaDataBase.append 
        """
        if not isinstance(dataframe, pd.DataFrame):
            raise TypeError("Argument dataframe is not of type Pandas.DataFrame")

        if tablename is None:
                tablename = self._get_valid_tablename(prefix="DATA_FRAME_")

        tablename = nzpyida.utils.check_tablename(tablename)

        if primary_key is not None:
            if not isinstance(primary_key, six.string_types):
                raise TypeError("The primary key argument should be a string")
            if primary_key not in dataframe.columns:
                raise ValueError("The primary key should be the name of a column" +
                                 " in the given dataframe")
            if len(dataframe[primary_key]) != len(set(dataframe[primary_key])):
                raise PrimaryKeyError(primary_key + " cannot be a primary key for" +
                                      "table " + tablename + " because it contains" +
                                      " non unique values")

        if self.exists_table_or_view(tablename):
            if clear_existing:
                try:
                    self.drop_table(tablename)
                except:
                    self.drop_view(tablename)
            else:
                raise NameError(("%s already exists, choose a different name "+
                                "or use clear_existing option.")%tablename)

        self._create_table(dataframe, tablename, primary_key=primary_key)
        idadf = nzpyida.frame.IdaDataFrame(self, tablename, indexer)
        self.append(idadf, dataframe)

        ############## Experimental ##################
        # dataframe.to_sql(tablename, self._con, index=False)
        # idadf = ibmdbpy.frame.IdaDataFrame(self, tablename, flavor='mysql',
        #                                    schema = self.current_schema)

        self._autocommit()

        if primary_key:
            idadf._indexer=primary_key
        return idadf
        
    ###########################################################################
    #### Delete DataBase objects
    ###########################################################################

    def drop_table(self, tablename):
        """
        Drop a table in the database.

        Parameters
        ----------
        tablename : str
            Name of the table to drop.

        Raises
        ------
        ValueError
            If the object does not exist.
        TypeError
            If the object is not a table.

        Examples
        --------
        >>> idadb.drop_table("TABLE")
        True
        >>> idadb.drop_table("NO_TABLE")
        TypeError : NO_TABLE  exists in schema '?' but of type '?'
        >>> idadb.drop_table("NOT_EXISTING")
        ValueError : NO_EXISTING doesn't exist in database

        Notes
        -----
            This operation cannot be undone if autocommit mode is activated.
        """
        return self._drop(tablename, "T")

    def drop_view(self, viewname):
        """
        Drop a view in the database.

        Parameters
        ----------
        viewname : str
            Name of the view to drop.

        Raises
        ------
        ValueError
            If the object does not exist.
        TypeError
            If the object is not a view.

        Examples
        --------
        >>> idadb.drop_view("VIEW")
        True
        >>> idadb.drop_view("NO_VIEW")
        TypeError : NO_VIEW exists in schema '?' but of type '?'
        >>> idadb.drop_view("NOT_EXISTING")
        ValueError : NO_EXISTING doesn't exist in database

        Notes
        -----
            This operation cannot be undone if autocommit mode is activated.
        """
        return self._drop(viewname, "V")

    def drop_model(self, modelname):
        """
        Drop a model in the database.

        Parameters
        ----------
        modelname : str
            Name of the model to drop.

        Raises
        ------
        ValueError
            If the object does not exist.
        TypeError
            if the object exists but is not a model.

        Examples
        --------
        >>> idadb.drop_model("MODEL")
        True
        >>> idadb.drop_model("NO_MODEL")
        TypeError : NO_MODEL exists in schema '?' but of type '?'
        >>> idadb.drop_model("NOT_EXISTING")
        ValueError : NOT_EXISTING does not exist in database

        Notes
        -----
            This operation cannot be undone if autocommit mode is activated.

        """

        try:
            self._call_stored_procedure("DROP_MODEL ", model = modelname)
        except Exception as e:
            try:
                flag = self.exists_table(modelname)
            except TypeError:
                # Exists but is not a model (nor a table)
                raise
            else:
                if flag:
                    # It is a table so make it raise by calling exists_view
                    self.exists_view(modelname)
            value_error = ValueError(modelname + " does not exist in database")
            six.raise_from(value_error, e)
        else:
            tables = self.show_tables()
            if not tables.empty:
                for table in tables['TABNAME']:
                    if modelname in table:
                        self.drop_table(table)
            return True

    @timed
    def rename(self, idadf, newname):
        """
        Rename a table referenced by an IdaDataFrame in Db2 Warehouse.

        Parameters
        ----------
        idadf : IdaDataFrame
            IdaDataFrame object referencing the table to rename.
        newname : str
            Name to be given to self. It should contain only alphanumeric
            characters and underscores. All lower case characters will be 
            converted to upper case characters. The new name should not already 
            exist in the database.

        Raises
        ------
        ValueError
            The new tablename is not valid.
        TypeError
            Rename function is supported only for table type.
        NameError
            The name of the object to be created is identical to an existing name.

        Notes
        -----
            Upper case characters and numbers, optionally separated by 
            underscores “_”, are valid characters.
        """
        # Actually we could support it for views too
        # Question : Is it better to accept an idadf as argument or rather the name of the table?
        oldname = idadf._name
        newname = nzpyida.utils.check_tablename(newname)

        if self.is_table(idadf._name):
            if self._is_netezza_system():
                query = "ALTER TABLE %s RENAME TO %s"%(idadf._name, newname)
            else:
                query = "RENAME TABLE %s TO %s"%(idadf._name, newname)
            try:
                self._prepare_and_execute(query)
            except Exception as e:
                if self._con_type == "odbc":
                    raise NameError(e.value[-1])
                else:
                    if self._is_netezza_system():
                        if "ERROR:  relation does not exist" in str(e.args[0]):
                            raise ValueError("Object does not exist.")
                        elif 'ERROR:  ALTER TABLE: object "%s" already exists'%newname in str(e.args[0]):
                            raise NameError("The new name is identical to the old one")
                        else:
                            raise e
                    else:
                        sql_code = int(str(e.args[0]).split("SQLCODE=")[-1].split(",")[0])
                        if sql_code == -601:
                            raise NameError("The new name is identical to the old one")
                        else:
                            raise e

            idadf._name = newname
            idadf.tablename = newname
            idadf.internal_state.name = newname
            self._reset_attributes("cache_show_tables")

            # Update name of all IdaDataFrame that were opened on this table
            for idadf in self._idadfs:
                if idadf._name == oldname:
                    idadf._name = newname
                    idadf.internal_state.name = newname # to refactor
        else:
            raise TypeError("Rename function is supported only for table type")

    def add_column(self, idadf, column = None, colname = None, ncat=10):
        """
        Add physically a column to the dataset.
        If column argument is set to None, a random column is generated which 
        can take values for 1 to ncat.
        Used for benchmark purpose. 
        Note: This method is deprecated. Can probably be removed without impact 
        """
        if column is not None:
            if column not in idadf.columns:
                raise ValueError("Unknown column %s"%column)
            dtype = idadf.dtypes['TYPENAME'][column]
            if dtype == "VARCHAR":
                dtype = dtype + "(255)"
        else:
            dtype = "INTEGER"
        
        if colname is None:
            i = 0
            colname = "COL%s"%i
            while colname in idadf.columns:
                i += 1
                colname = "COL%s"%i
        else:
            if colname in idadf.columns:
                raise ValueError("Column %s already exists"%colname)
        
        self.commit() 
        query = "ALTER TABLE %s ADD \"%s\" %s"%(idadf._name, colname, dtype) 
        print(query)         
        try:
            self._prepare_and_execute(query)
        except:
            print("Failed ! Trying again in 5 seconds")
            #query2 = "REORG TABLE %s"%idadf._name
            #print(query2)
            import time
            time.sleep(5)
            self._prepare_and_execute(query)
        self.commit()
        
        # force sleep 2 sec
        
        
        if column is None:
            # Random column
            query = "UPDATE %s SET \"%s\" = RAND()* %s + 1"%(idadf._name, colname, ncat)
        else:
            query = "UPDATE %s SET \"%s\" = \"%s\""%(idadf._name, colname, column)
        
        print(query)
        
        try:
            self._prepare_and_execute(query)
        except:
            print("Failed ! Trying again in 5 seconds")
            import time
            time.sleep(5)
            self._prepare_and_execute(query)
                
            
        self.commit()
        idadf._reset_attributes(["columns", "dtypes", "shape"])
        
            
    
    @timed
    def add_column_id(self, idadf, column_id="ID", destructive=False):
        # TODO: Base the creation of the idea on the sorting of several columns
        # (or all columns in case there are duplicated rows) so that the ID
        # can be created in a determinstic and reproducible way
        """
        Add an ID column to an IdaDataFrame.

        Parameters
        ----------
        idadf : IdaDataFrame
            IdaDataFrame object to which an ID column will be added
        column_id : str
            Name of the ID column to add
        destructive : bool
            If set to True, the column will be added phisically in the database.
            This can take time. If set to False, the column will be added virtually
            in a view and a new IdaDataFrame is returned.

        Raises
        ------
        TypeError
            idadf is not an IdaDataFrame.
        ValueError
            The given column name already exists in the DataBase.

        Notes
        -----
            The non-destructive creation of column IDs is not reliable, because row IDs are recalculated on the fly in a non-deterministic 
            way each time a new view is produced. On the contrary, creating them destructively i.e physically is reliable but can take time. 
            If no sorting has been done whatsoever before, row IDs will be created at random.
            Improvement idea: create ID columns in a non-destructive way and base them on the sorting of a set of columns, 
            defined by the user, or all columns if no column combination results in unique identifiers.
        """
        if isinstance(idadf, nzpyida.IdaSeries):
            raise TypeError("Adding column ID is not supported for IdaSeries")
        if not isinstance(idadf, nzpyida.IdaDataFrame):
            raise TypeError("idadf is not an IdaDataFrame, type: %s"%type(idadf))
        if column_id in idadf.columns:
            raise ValueError("A column named '"+column_id+"' already exists." +
                             " Please define a new column name using"+
                             " column_id argument")

        if destructive is True:

            viewname = self._get_valid_tablename(prefix="VIEW_")
            if self._is_netezza_system():
                order_by = "ORDER BY NULL"
            else:
                order_by = ""
            self._prepare_and_execute("CREATE VIEW " + viewname + " AS SELECT ((ROW_NUMBER() OVER("+ order_by +"))-1)" +
                                      " AS \"" + column_id + "\", \""+
                                      "\",\"".join(idadf._get_all_columns_in_table()) +
                                      "\" FROM " + idadf._name)

            # Initiate the modified table under a random name
            tablename = self._get_valid_tablename(prefix="DATA_FRAME_")
            if self._is_netezza_system():
                 self._prepare_and_execute("CREATE TABLE %s AS (SELECT * FROM %s)"%(tablename,viewname))
            else:
                self._prepare_and_execute("CREATE TABLE %s LIKE %s"%(tablename,viewname))
                self._prepare_and_execute("INSERT INTO %s (SELECT * FROM %s)"%(tablename,viewname))

            # Drop the view and old table
            self.drop_view(viewname)
            self.drop_table(idadf._name)

            # Give it the original name back
            self._reset_attributes("cache_show_tables")
            new_idadf = nzpyida.IdaDataFrame(self, tablename)
            self.rename(new_idadf, idadf.tablename)

            # Updating internal state
            # prepend the columndict OrderedDict
            items = idadf.internal_state.columndict.items()
            idadf.internal_state.columndict = OrderedDict()
            idadf.internal_state.columndict[column_id] = "\"" + column_id + "\""
            for item in items:
                idadf.internal_state.columndict[item[0]] = item[1]

            idadf.internal_state.update()
            idadf._reset_attributes(["columns"])
            idadf.indexer = column_id
        else:
            # prepend the columndict OrderedDict
            items = idadf.internal_state.columndict.items()
            idadf.internal_state.columndict = OrderedDict()
            if self._is_netezza_system():
                order_by = "ORDER BY NULL"
            else:
                order_by = ""
            idadf.internal_state.columndict[column_id] = "((ROW_NUMBER() OVER("+ order_by +"))-1)"
            for item in items:
                idadf.internal_state.columndict[item[0]] = item[1]
            idadf.internal_state.update()
            idadf._reset_attributes(["columns"])
            idadf.indexer = column_id

        # Reset attributes
        idadf._reset_attributes(['shape', 'columns', 'axes', 'dtypes'])

    def delete_column(self, idadf, column_name, destructive=False):
        """
        Delete a column in an idaDataFrame.

        Parameters
        ----------
        idadf : IdaDataFrame
            The IdaDataframe in which a column should be deleted.
        column_name : str
            Name of the column to delete.
        destructive : bool
            If set to True, the column is deleted in the database. Otherwise, 
            it is deleted virtually, creating a view for the IdaDataFrame.

        Raises
        ------
        TypeError
            column_name should be a string.
        ValueError
            column_name refers to a column that doesn't exist in self.
        """
        if not isinstance(column_name, six.string_types):
            raise TypeError("column_name is not of string type")
        if column_name not in idadf.columns:
            raise ValueError("%s refers to a columns that doesn't exists in self"%(column_name))

        if destructive is True:
            if column_name not in idadf._get_all_columns_in_table():
                # Detect it is a virtual ID, the deletion cannot be destructive
                return self.delete_column(idadf, column_name, destructive=False)

            viewname = self._get_valid_tablename(prefix="VIEW_")
            columnlist = list(idadf._get_all_columns_in_table())
            columnlist.remove(column_name)

            self._prepare_and_execute("CREATE VIEW " + viewname + " AS SELECT \""+
                                      "\",\"".join(columnlist) +
                                      "\" FROM " + idadf._name)

            tablename = self._get_valid_tablename(prefix="DATA_FRAME_")

            if self._is_netezza_system():
                self._prepare_and_execute("CREATE TABLE %s AS (SELECT * FROM %s)"%(tablename,viewname))
            else:
                self._prepare_and_execute("CREATE TABLE " + tablename + " LIKE " + viewname)
                self._prepare_and_execute("INSERT INTO " + tablename + " (SELECT * FROM " + viewname + ")")

            # Drop the view and old table
            self.drop_view(viewname)
            self.drop_table(idadf._name)

            # Give it the original name back
            self._reset_attributes("cache_show_tables") # normally, no needed
            new_idadf = nzpyida.IdaDataFrame(self, tablename, idadf.indexer)
            self.rename(new_idadf, idadf.tablename)

            # updating internal state
            del idadf.internal_state.columndict[column_name]
            idadf.internal_state.update()
            self._reset_attributes("cache_show_tables")
            idadf._reset_attributes(['shape', 'columns', 'dtypes'])

        else:
            del idadf.internal_state.columndict[column_name]
            idadf.internal_state.update()
            idadf._reset_attributes(["columns", "shape", "dtypes"])

        if column_name == idadf.indexer:
            idadf._reset_attributes(["_indexer"])


    def append(self, idadf, df, maxnrow=None):
        """
        Append rows of a DataFrame to an IdaDataFrame. The DataFrame must have 
        the same structure (same column names and datatypes). Optionally, the 
        DataFrame to be added can be splitted into several chunks. This 
        improves performance and prevents SQL overflows. By default, chunks are 
        limited to 100.000 cells.

        Parameters
        ----------
        idadf : IdaDataFrame
            IdaDataFrame that receives data from dataframe df.
        df : DataFrame
            Dataframe whose rows are added to IdaDataFrame idadf.
        maxnrow : int, optional
            number corresponding to the maximum number of rows for each chunks.

        Raises
        ------
        TypeError
            * maxnrow should be an interger.
            * Argument idadf should be an IdaDataFrame.
            * Argument df should be a pandas DataFrame.
        ValueErrpr
            * maxnrow should be greater than 1 or nleft blank.
            * Other should be a Pandas DataFrame.
            * Other dataframe has not the same number of columns as self.
            * Some columns in other have different names that are different from the names of the columns in self.
        """
        # SANITY CHECK : maxnrow
        if maxnrow is None:
            # Note : it has been measured on a big dataset (>1 million rows) that int(100000 / len(df.columns)) 
            # performs better than the previous empirical value int(8000 / len(df.columns)) 
            maxnrow = int(100000 / len(df.columns))
        else:
            if not isinstance(maxnrow, six.integer_types):
                raise TypeError("maxnrow is not an integer")
            if maxnrow < 1:
                raise ValueError("maxnrow should be stricly positive or omitted")
            if maxnrow > 15000:
                warnings.warn("Performance may decrease if maxnrow is bigger than 15000", UserWarning)

        # SANITY CHECK : idadf & other
        if not isinstance(idadf, nzpyida.frame.IdaDataFrame):
            raise TypeError("Argument idadf is not an IdaDataFrame")
        if not isinstance(df, pd.DataFrame):
            raise TypeError("Argument df is not of type pPandas.DataFrame")
        if len(df.columns) != len(idadf.columns):
            raise ValueError("(Ida)DataFrames don't have the same number of columns")
        if any([column not in idadf.columns for column in [str(x).strip() for x in df.columns]]):
            raise ValueError("Some column names do not match current object columns: \n" +
                             "Expected : \t" + str(idadf.columns) + "\n" +
                             "Found : \t" + str([x.strip() for x in df.columns]) + "\n")
        if any([str(column_idadf) != str(column_other).strip() for column_idadf, column_other in
                zip(idadf.columns, df.columns)]):
            raise ValueError("Order or columns in other and" + idadf._name + "does not match.")

        if df.shape[0] > 1.5 * maxnrow:
            split_into = math.ceil(df.shape[0] / maxnrow)
            split = np.array_split(df, split_into)
            print("DataFrame will be splitted into " + str(split_into) +
                  " chunks. (" + str(maxnrow) + " rows per chunk)")
            for i, chunk in enumerate(split, 0):
                percentage = int(i / split_into * 100)
                print("Uploaded: " + str(percentage) + "%... ", end="\r")
                try:
                    self._insert_into_database(chunk, idadf.schema, idadf.tablename, silent=True)
                except:
                    raise
            print("Uploaded: %s/%s... "%(split_into,split_into), end="")
            print("[DONE]")
        else:
            print("Uploading %s rows (maxnrow was set to %s)"%(df.shape[0], maxnrow))
            try:
                self._insert_into_database(df, idadf.schema, idadf.tablename, silent=True)
            except:
                raise

        idadf._reset_attributes(['shape', 'axes', 'dtypes', 'index'])

    def merge(self, idadf, other, key):
        # TODO:
        pass

    ###############################################################################
    #### Connection management
    ###############################################################################

    def commit(self):
        """
        Commit operations in the database.

        Notes
        -----

        All changes that are made in the database after the last commit, 
        including those in the child IdaDataFrames, are commited.

        If the environment variable ‘VERBOSE’ is set to True, the commit 
        operations are notified in the console.
        """
        self._check_connection()  # Important
        self._con.commit()
        if os.getenv('VERBOSE') == 'True':
            print("<< COMMIT >>")
        self._reset_attributes("cache_show_tables")

    def rollback(self):
        """
        Rollback operations in the database.

        Notes
        -----

        All changes that are made in the database after the last commit, 
        including those in the child IdaDataFrames, are discarded.
        """
        self._check_connection()  # Important
        self._con.rollback()
        if os.getenv('VERBOSE') == 'True':
            print("<< ROLLBACK >>")
        self._reset_attributes("cache_show_tables")

    def close(self):
        """
        Close the IdaDataBase connection.

        Notes
        -----

        If the environment variable ‘AUTOCOMMIT’ is set to True, then all 
        changes after the last commit are committed, otherwise they are 
        discarded.

        """
        if os.getenv('AUTOCOMMIT') == 'True':
            self.commit()
        else:
            self.rollback()
        self._reset_attributes("cache_show_tables")
        self._con.close()
        print("Connection closed.")

    def reconnect(self):
        """
        Try to reopen the connection.
        """
        try:
            self._check_connection()
        except IdaDataBaseError:
            if self._con_type == 'odbc':
                import pyodbc
                try:
                    self._con = pyodbc.connect(self._connection_string)
                except:
                    raise
                else:
                    print("The connection was successfully restored")
            elif self._con_type == 'nzpy':
                import nzpy
                try:
                    self._con = nzpy.connect(**(self._connection_string))
                except:
                    raise
                else:
                    print("The connection was successfully restored")
            elif self._con_type == 'jdbc':
                try:
                    import jaydebeapi
                    if self._is_netezza_system():
                        self._con = jaydebeapi.connect('org.netezza.Driver', self._connection_string)
                    else:
                        self._con = jaydebeapi.connect('com.ibm.db2.jcc.DB2Driver', self._connection_string)
                except:
                    raise
                else:
                    print("The connection was successfully restored")
        else:
            print("The connection for current IdaDataBase is valid")

        ###############################################################################
        #### Private methods
        ###############################################################################

    def __enter__(self):
        """
        Allow the object to be used with a "with" statement
        """
        return self

    def __exit__(self):
        """
        Allow the object to be used with a "with" statement. Make sure the
        connection is closed when the object get out of scope
        """
        self.close()

    def _exists(self, objectname, typelist):
        """
        Check if an object of a certain type exists in Db2 Warehouse.

        Notes
        -----
        For more information, see exists_table_or_view, exists_table, 
        exists_view functions.
        """
        objectname = nzpyida.utils.check_tablename(objectname)

        tablelist = self.show_tables(show_all=True)
        schema, name = self._get_name_and_schema(objectname)
        tablelist = tablelist[tablelist['TABSCHEMA'] == schema]

        if len(tablelist):
            if name in tablelist['TABNAME'].values:
                tabletype = tablelist[tablelist['TABNAME'] == name]['TYPE'].values[0]
                if tabletype in typelist:
                    return True
                else:
                    raise TypeError("%s exists in schema %s but of type \"%s\""
                                    %(objectname,schema,tabletype))
            else:
                return False
        else:
            return False

    def _is(self, objectname, typelist):
        """
        Check if an existing object is of a certain type or in a list of types.

        Notes
        -----
        For more information, see is_table_or_view, is_table, is_view functions.
        """
        objectname = nzpyida.utils.check_tablename(objectname)

        tablelist = self.show_tables(show_all=True)
        schema, name = self._get_name_and_schema(objectname)
        tablelist = tablelist[tablelist['TABSCHEMA'] == schema]

        if len(tablelist):
            if name in tablelist['TABNAME'].values:
                tabletype = tablelist[tablelist['TABNAME'] == name]['TYPE'].values[0]
                if tabletype in typelist:
                    return True
                else:
                    return False

        raise ValueError("%s does not exist in database"%(objectname))

    def _drop(self, objectname, object_type = "T"):
        """
        Drop an object in the table depending on its type.
        Admissible type values are "T" (table) and "V" (view)

        Notes
        -----
        For more information, seedrop_table and drop_view functions.

        """
        objectname = nzpyida.utils.check_tablename(objectname)

        if object_type == "T":
            to_drop = "TABLE"
        elif object_type == "V":
            to_drop = "VIEW"
        else:
            raise ValueError("Unknown type to drop")

        try:
            self._prepare_and_execute("DROP %s %s"%(to_drop,objectname))
        except Exception as e:
            if self._con_type == "odbc":
                if e.value[0] == "42S02":
                    raise ValueError(e.value[1])  # does not exist
                if e.value[0] == "42809":
                    raise TypeError(e.value[1])  # object is not of expected type
            else:
                if self._is_netezza_system():
                    if "ERROR:  relation does not exist" in str(e.args[0]):
                        raise ValueError("Object does not exist.")
                    else:
                        raise e
                else:
                    sql_code = int(str(e.args[0]).split("SQLCODE=")[-1].split(",")[0])
                    if sql_code == -204:
                        raise ValueError("Object does not exist.")
                    elif sql_code == -159:
                        raise TypeError("Object is not of expected type")
                    else:
                        raise e # let the expection raise anyway
        else:
            self._reset_attributes("cache_show_tables")
            return True

    def _upper_columns(self, dataframe):
        # Could be moved to utils (then move in the test too)
        """
        Put every column name of a Pandas DataFrame in upper case.

        Parameters
        ----------
        dataframe : DataFrame

        Returns
        -------
        DataFrame
        """
        data = deepcopy(dataframe)
        if len(data):
            data.columns = [x.upper() for x in data.columns]
        return data

    def _get_name_and_schema(self, objectname):
        """
        Helper function that returns the name and the schema from an object 
        name. Implicitly, if no schema name was given, it is assumed that user 
        refers to the current schema.

        Parameters
        ----------
        objectname : str
            Name of the object to process. Can be either under the form
            "SCHEMA.TABLE" or just "TABLE"

        Returns
        -------
        tuple
            A tuple composed of 2 strings containing the schema and the name.

        Examples
        --------
        >>> _get_name_and_schema(SCHEMA.TABLE)
        (SCHEMA, TABLE)
        >>> _get_name_and_schema(TABLE)
        (<current schema>, TABLE)
        """
        if '.' in objectname:
            name = objectname.split('.')[-1]
            schema = objectname.split('.')[0]
        else:
            name = objectname
            schema = self.current_schema
        return (schema, name)

    def _get_valid_tablename(self, prefix="DATA_FRAME_"):
        """
        Generate a valid database table name.

        Parameters
        ----------
        prefix : str, default: "DATA_FRAME_"
            Prefix used to create the table name. The name is constructed using 
            this pattern : <prefix>_X where <prefix> corresponds to the string 
            parameter “prefix” capitalized and X corresponds to a pseudo 
            randomly generated number (0-100000).

        Returns
        -------
        str

        Examples
        --------
        >>> idadb._get_valid_tablename()
        'DATA_FRAME_49537_1434978215'
        >>> idadb._get_valid_tablename("MYDATA_")
        'MYDATA_65312_1434978215'
        >>> idadb._get_valid_tablename("mydata_")
        'MYDATA_78425_1434978215'
        >>> idadb._get_valid_tablename("mydata$")
        ValueError: Table name is not valid, only alphanumeric characters and underscores are allowed.
        """
        prefix = nzpyida.utils.check_tablename(prefix)
        # We may assume that the generated name is unlikely to exist
        name = "%s%s_%s" % (prefix, random.randint(0, 100000), int(time()))
        return name

    def _get_valid_viewname(self, prefix="VIEW_"):
        """
        Convenience function : Alternative name for _get_valid_tablename.

        The parameter prefix has its optional value changed to "VIEW\_"".

        Examples
        --------
        >>> idadb._get_valid_viewname()
        'VIEW_49537_1434978215'
        >>> idadb._get_valid_viewname("MYVIEW_")
        'MYVIEW_65312_1434978215'
        >>> idadb._get_valid_viewname("myview_")
        'MYVIEW_78425_1434978215'
        >>> idadb._get_valid_modelname("myview$")
        ValueError: View name is not valid, only alphanumeric characters and underscores are allowed.
        """
        return self._get_valid_tablename(prefix)

    def _get_valid_modelname(self, prefix="MODEL_"):
        """
        Convenience function : Alternative name for _get_valid_tablename.

        Parameter prefix has its optional value changed to "MODEL\_".

        Examples
        --------
        >>> idadb._get_valid_modelname()
        'MODEL_49537_1434978215'
        >>> idadb._get_valid_modelname("TEST_")
        'TEST_65312_1434996318'
        >>> idadb._get_valid_modelname("test_")
        'TEST_78425_1435632423'
        >>> idadb._get_valid_tablename("mymodel$")
        ValueError: Table name is not valid, only alphanumeric characters and underscores are allowed.
        """
        return self._get_valid_tablename(prefix)

    def _create_table(self, dataframe, tablename, primary_key=None):
        """
        Create a new table in the database by declaring its name and columns 
        based on an existing DataFrame. It is possible declare a column as 
        primary key.

        Parameters
        ----------
        dataframe : DataFrame
            Pandas DataFrame be used to initiate the table.
        tablename : str
            Name to be given to the table at its creation.
        primary_key: str
            Name of a column to declare as primary key.

        Notes
        -----
        The columns and their data type is deducted from the Pandas DataFrame 
        given as parameter.

        Examples
        --------
        >>> from nzpyida.sampledata.iris import iris
        >>> idadb._create_table(iris, "IRIS")
        'IRIS'
        >>> idadb._create_table(iris)
        'DATA_FRAME_4956'
        """
        # TODO : Handle more types
        # integer_attributes =
        # ['int_', 'intc','BIGINT','REAL','DOUBLE','FLOAT','DECIMAL','NUMERIC']
        # double_attributes =
        # ['SMALLINT', 'INTEGER','BIGINT','REAL','DOUBLE','FLOAT','DECIMAL','NUMERIC']
        # self._check_connection()

        if not isinstance(dataframe, pd.DataFrame):
            raise TypeError("_create_table is valid only for DataFrame objects")

        if primary_key is not None:
            if not isinstance(primary_key, six.string_types):
                raise TypeError("primary_key argument should be a string")

        # Check the tablename
        tablename = nzpyida.utils.check_tablename(tablename)

        # for Netezza we have to check if the schema exists already
        # otherwise we have to create it
        if self._is_netezza_system() & ("." in tablename):
            schemaname, tabname = tablename.split(".")
            if self.ida_scalar_query("SELECT COUNT(*) FROM _V_SCHEMA WHERE SCHEMA='%s'"%schemaname) == 0:
                self.ida_query("CREATE SCHEMA " + schemaname)

        column_string = ''
        for column in dataframe.columns:
            if dataframe.dtypes[column] in [object,bool]:
                # Handle boolean type
                if set(dataframe[column].unique()).issubset([True, False, 0, 1, np.nan]):
                    column_string += "\"%s\" SMALLINT," % str(column).strip()
                else:
                    if column == primary_key:
                        column_string += "\"%s\" VARCHAR(255) NOT NULL, PRIMARY KEY (\"%s\")," % (str(column).strip(), str(column).strip())
                    else:
                        column_string += "\"%s\" VARCHAR(255)," % str(column).strip()
            elif dataframe.dtypes[column] == np.dtype('datetime64[ns]'):
                # This is a first patch for handling dates
                # TODO: Dates as timestamp in the database
                column_string += "\"%s\" VARCHAR(255)," % str(column).strip()
            else:
                if dataframe.dtypes[column] in [np.int64, np.int, np.int8, np.int32]:
                    if abs(dataframe[column].max()) < 2147483647/2: # might get bigger 
                        datatype = "INT"
                    else:
                        datatype = "BIGINT"
                else:
                    datatype = "DOUBLE"
                if column == primary_key:
                    column_string += "\"%s\" %s NOT NULL, PRIMARY KEY (\"%s\")," % (str(column).strip(), datatype, str(column).strip())
                else:
                    column_string += "\"%s\" %s," %(str(column.strip()), datatype)
        if column_string[-1] == ',':
            column_string = column_string[:-1]

        if '.' in tablename:
            schema, tablename2 = tablename.split('.')
        else:
            schema = self.current_schema
            tablename2 = tablename

        create_table = "CREATE TABLE \"%s\".\"%s\" (%s)" % (schema, tablename2, column_string)

        self._prepare_and_execute(create_table, autocommit=False)

        # Save new table in cache
        if hasattr(self, "cache_show_tables"):
            record = (schema, tablename2, self.current_schema, 'T')
            self.cache_show_tables.loc[len(self.cache_show_tables)] = np.array(record)

        return "\"%s\".\"%s\"" % (schema, tablename2)

    def _create_view(self, idadf, viewname = None):
        """
        Create a new view in the database from an existing table.

        Parameters
        ----------

        idadf : IdaDataFrame
            IdaDataFrame to be duplicated as view.

        viewname : str, optional
            Name to be given to the view at its creation. If not given, a 
            random name will be generated automatically 

        Returns
        -------
        str
            View name.

        Examples
        --------
        >>> idadf = IdaDataFrame(idadb, "IRIS")
        >>> idadb._create_view(idadf)
        'IDAR_VIEW_4956'
        """
        if not isinstance(idadf, nzpyida.frame.IdaDataFrame):
            raise TypeError("_create_view is valid only for IdaDataFrame objects")

        # Check the viewname
        if viewname is not None:
            viewname = nzpyida.utils.check_viewname(viewname)
        else:
            viewname = self._get_valid_viewname()

        self._prepare_and_execute("CREATE VIEW \"%s\" AS SELECT * FROM %s" % (viewname, idadf._name))
    
        # Save new view in cache
        if hasattr(self, "cache_show_tables"):
            record = (self.current_schema, viewname, self.current_schema, 'V')
            self.cache_show_tables.loc[len(self.cache_show_tables)] = np.array(record)
    
        return viewname
        
    def _create_view_from_expression(self, expression, viewname = None):
        """
        Create a new view in the database from an expression.

        Parameters
        ----------

        expression : str
            Expression that defines the view to be created

        viewname : str, optional
            Name to be given to the view at its creation. If not given, a 
            random name will be generated automatically 

        Returns
        -------
        str
            View name.

        Examples
        --------
        TODO
        """
        if not isinstance(expression, six.string_types):
            raise TypeError("expression argument expected to be of string type")

        # Check the viewname
        if viewname is not None:
            viewname = nzpyida.utils.check_viewname(viewname)
        else:
            viewname = self._get_valid_viewname()

        self._prepare_and_execute("CREATE VIEW \"%s\" AS %s" % (viewname, expression))
    
        # Save new view in cache
        if hasattr(self, "cache_show_tables"):
            record = (self.current_schema, viewname, self.current_schema, 'V')
            self.cache_show_tables.loc[len(self.cache_show_tables)] = np.array(record)
    
        return viewname

    def _insert_into_database(self, dataframe, schema, tablename, silent=True):
        """
        Populate an existing table with data from a dataframe.

        Parameters
        ----------
        dataframe: DataFrame
            Data to be inserted into an existing table, contained in a Pandas 
            DataFrame. It is assumed that the structure matches.
        schema: str
            Schema of the table in which the data is inserted.
        tablename: str
            Name of the table in which the data is inserted.
        silent : bool, default: True
            If True, the INSERT statement is not printed. Avoids flooding the 
            console.

        """
        # TODO : Handle more datatypes
        if schema is None or schema.strip() == '':
            schema = self.current_schema
        tablename = nzpyida.utils.check_tablename(tablename)
        column_string = '\"%s\"' % '\", \"'.join([str(x).strip() for x in dataframe.columns])
        row_string = ''

        # Save in a list columns that are booleans
        boolean_flaglist = []
        for column in dataframe.columns:
            if set(dataframe[column].unique()).issubset([True, False, 0, 1, np.nan]):
                boolean_flaglist.append(1)
            else:
                boolean_flaglist.append(0)

        if self._is_netezza_system():
            sepstr1 = 'SELECT '
            sepstr2 = ' UNION ALL SELECT '
            sepstr3 = ' '
        else:
            sepstr1 = 'VALUES ('
            sepstr2 = '), ('
            sepstr3 = ') '

        row_separator = sepstr1
        for rows in dataframe.values:
            value_string = ''
            for colindex, value in enumerate(rows):
                if pd.isnull(value): # handles np.nan and None
                    value_string += "NULL,"  # Handle missing values
                elif isinstance(value, six.string_types):
                    ## Handle apostrophe in values
                    value = value.replace("\\", "'")
                    value_string += '\'%s\',' % value.replace("'", "''")
                # REMARK: it is the best way to handle booleans ?
                elif isinstance(value, bool):
                    if boolean_flaglist[colindex] == True:
                        if value in [1, True]:
                            value_string += '1,'
                        elif value in [0, False]:
                            value_string += '0,'
                    else:
                        value_string += '\'%s\',' % value
                # TODO: Handle datetime better than strings
                elif isinstance(value, datetime.datetime):
                    value_string += '\'%s\',' % value
                else:
                    value_string += '%s,' % value
            if value_string[-1] == ',':
                value_string = value_string[:-1]
            row_string += (row_separator + " %s ") % value_string
            row_separator = sepstr2
        row_string = row_string +  sepstr3
        if row_string[-2:] == '),':
            row_string = row_string[:-2]
        if row_string[0] == '(':
            row_string = row_string[1:]
        query = ("INSERT INTO \"%s\".\"%s\" (%s) (%s)" % (schema, tablename, column_string, row_string))

        # print(query)
        # TODO: Good idea : create a savepoint before creating the table
        # Rollback in to savepoint in case of failure
        self._prepare_and_execute(query, autocommit=False, silent=silent)

        for idadf in self._idadfs:
            if idadf._name == tablename:
                idadf._reset_attributes(["shape", "index"])

    def _prepare_and_execute(self, query, autocommit=True, silent=False):
        """
        Prepare and execute a query by using the cursor of an idaobject.

        Parameters
        ----------
        idaobject: IdaDataBase or IdaDataFrame
        query: str
            Query to be executed.
        autocommit: bool, default: True
            If True, the autocommit function is available.
        silent: bool, default: False
            If True, the SQL statement is not printed.
        """
        self._check_connection()
        return sql._prepare_and_execute(self, query, autocommit, silent)

    def _check_procedure(self, proc_name, alg_name=None):
        """
        Check if a procedure is available in the database.

        Parameters
        ----------
        proc_name : str
            Name of the procedure to be checked as defined in the underlying
            database management system.
        alg_name : str
            Name of the algorithm, human readable.

        Returns
        -------
        bool

        Examples
        --------
        >>> idadb._check_procedure('KMEANS')
        True
        >>> idadb._check_procedure('NOT_EXISTING')
        IdaDatabaseError: Function 'NOT_EXISTING' is not available.
        """

        if alg_name is None:
            alg_name = proc_name
        if self._is_netezza_system():
            query = ("SELECT COUNT(*) FROM NZA.._V_PROCEDURE WHERE PROCEDURE='%s'") % proc_name
        else:
            query = ("SELECT COUNT(*) FROM SYSCAT.ROUTINES WHERE ROUTINENAME='%s" +
                     "' AND ROUTINEMODULENAME = 'IDAX'") % proc_name
        flag = self.ida_scalar_query(query)

        if int(flag) == False:
            raise IdaDataBaseError("Function '%s' is not available." % alg_name)
        else:
            return True

    def _call_stored_procedure(self, sp_name, **kwargs):
        """
        Call a specific IDAX/INZA stored procedure and return its result.

        Parameters
        ----------
        sp_name : str
            Name of the stored procedure.
        **kwargs : ...
            Additional parameters, specific to the called stored procedure.
        """
        tmp = []
        views = []
        if self._is_netezza_system():
            sp_schema = 'NZA.'
        else:
            sp_schema = 'IDAX'

        for key, value in six.iteritems(kwargs):
            if value is None:
                continue  # Go to next iteration
            if isinstance(value, nzpyida.frame.IdaDataFrame):
                tmp_view_name = self._create_view(value)
                tmp.append("%s=%s" % (key, tmp_view_name))
                views.append(tmp_view_name)
            elif isinstance(value, six.string_types) and all([x != " " for x in value]):
                if key in ("intable", 'model', 'outtable', 'incolumn', 'nametable', 'colPropertiesTable'):
                    tmp.append("%s=%s"% (key, value))  # no " in the case it is a table name
                else:
                    tmp.append("%s=\"%s\"" % (key, value))
            elif isinstance(value, list):
                tmp.append("%s=\"%s\"" % (key, " ".join(str(value))))
            else:
                tmp.append("%s=%s" % (key, value))
        try:
            call = "CALL %s.%s('%s')" % (sp_schema, sp_name, ",".join(tmp))
            result = self._prepare_and_execute(call)
        except:
            if self._is_netezza_system():
                error_msg = "Error"
            else:
                error_msg = self.ida_scalar_query("values idax.last_message")
            raise IdaDataBaseError(error_msg)
        finally:
            for view in views:
                self.drop_view(view)

        return result

    def _autocommit(self):
        """
        Commit changes made to the database in the connection automatically.
        If the environment variable 'AUTOCOMMIT' is set to True, then commit.

        Notes
        -----
        In the case of a commit operation, all changes that are made in the
        Database after the last commit, including those in the children
        IdaDataFrames, will be commited.

        If the environment variable 'VERBOSE' is not set to 'True', the
        autocommit operations will not be notified in the console to the user.
        """
        if os.getenv('AUTOCOMMIT') == 'True':
            self._con.commit()
            if os.getenv('VERBOSE') == 'True':
                print("<< AUTOCOMMIT >>")

    def _check_connection(self):
        """
        Check if the connection still exists by trying to open a cursor.
        """
        if self._con_type == "odbc":
            try:
                self._con.cursor()
            except Exception as e:
                raise IdaDataBaseError(e.value[-1])
        elif self._con_type == "nzpy":
            try:
                self._con.cursor().execute("select 1")
            except Exception as e:
                raise IdaDataBaseError(e)
        elif self._con_type == "jdbc":
            # This avoids infinite recursions on Netezza due to reconnect calls in
            # sql.ida_query, sql.ida_scalar_query and sql._prepare_and_execute
            if self._con._closed:
                raise IdaDataBaseError("The connection is closed")
            try:
                # Avoid infinite recursion
                if self._is_netezza_system():
                    # On Netezza no result sets are returned after the first database error
                    if read_sql("SELECT OBJID FROM _V_TABLE LIMIT 1", self._con) is None:
                        raise IdaDataBaseError("The connection is closed")
                else:
                    read_sql("SELECT TABLEID FROM SYSCAT.TABLES LIMIT 1", self._con)
            except Exception as e:
                raise IdaDataBaseError("The connection is closed")

    def _retrieve_cache(self, cache):
        """
        Helper function that retrieve cache if available.
        Cache are just string type values stored in private attributes.
        """
        if not isinstance(cache, six.string_types):
            raise TypeError("cache is not of string type")

        self._check_connection()

        if hasattr(self, cache):
            return getattr(self,cache)
        else:
            return None

    def _reset_attributes(self, attributes):
        """
        Helper function that delete attributes given as parameter if they
        exists in self. This is used to refresh lazy attributes and caches.
        """
        nzpyida.utils._reset_attributes(self, attributes)


    def _is_netezza_system(self):
         """
         Checks if the underlying database system is Netezza.
         """
         return self._database_system == 'netezza'


# -*- coding: utf-8 -*-
"""
Created on Mon Dec 14 13:05:27 2015

@author: efouche
"""
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import absolute_import
from builtins import dict
from future import standard_library
standard_library.install_aliases()

from collections import OrderedDict

import numpy as np 
import pandas as pd

#from ibmdbpy.internals import idadf_state
from nzpyida.feature_selection.private import _check_input

#@idadf_state#(force=True)
def ttest(idadf, target=None, features=None, ignore_indexer=True):
    """
    Compute the t-statistics values of a set of features against a set of 
    target attributes. 
    
    Parameters
    ----------
    idadf : IdaDataFrame
    
    target : str or list of str, optional
        A column or list of columns against which the t-statistcs values will 
        be computed. Per default, consider all columns
    
    features : str or list of str, optional
        A column or list of columns for which the t-statistics values will be 
        computed against each target attributes. Per default, consider all 
        columns, except non numerical columns. 
    
    ignore_indexer : bool, default: True
        Per default, ignore the column declared as indexer in idadf
        
    Returns
    -------
    Pandas.DataFrame or Pandas.Series if only one target
    
    Raises
    ------
    TypeError
        If the features argument or the data set does not contains any 
        numerical features. Raise TypeError. 
        
    Notes
    -----
    This implements the "modified" ttest as defined in the paper
    A Modified T-test feature Selection Method and Its Application on
    the HapMap Genotype Data (Zhou et al.)
    
    The target columns should be categorical, while the feature columns should
    be numerical.
    
    The scalability of this approach is not very good. Should not be used on 
    high dimensional data. 
    
    Examples
    --------
    >>> idadf = IdaDataFrame(idadb, "IRIS")
    >>> ttest(idadf,"CLASS")
    """
    # Check input
    target, features = _check_input(idadf, target, features, ignore_indexer)
    ttest_dict = OrderedDict()
    length = len(idadf)
    
    S_dict = dict()
    M_dict = dict()
    class_mean_dict = dict()
    
    numerical_columns = idadf._get_numerical_columns()
    
    # Filter out non numerical columns
    features = [feature for feature in features if feature in numerical_columns]
    if not features:
        raise TypeError("No numerical features.")
        
    #mean = idadf[features].mean() # This is broken
    mean = idadf.mean()
    
    if target is None:
        target = list(idadf.columns)
            
    for t in target:
        features_notarget = [x for x in features if (x != t)]
    
        if t not in M_dict:
            count = idadf.count_groupby(t)    
            target_count = count["count"]
            target_count.index = count[t]
            M_dict[t] = np.sqrt(1/target_count + 1/length)     
        
        if t not in S_dict:
            S_dict[t] = idadf.within_class_std(target = t, features = features_notarget)
        
        if t not in class_mean_dict:
            class_mean_dict[t] = idadf.mean_groupby(t, features = features_notarget)
            
        M = M_dict[t]
        S = S_dict[t]
        class_mean = class_mean_dict[t]
        
        ttest_dict[t] = OrderedDict()
        for feature in features_notarget:
            ttest_dict[t][feature] = OrderedDict()
            for target_class in class_mean.index:
                numerator = abs(class_mean.loc[target_class][feature] - mean[feature])
                denominator = M[target_class] * S[feature]
                
                ttest_dict[t][feature][target_class] = numerator / denominator
                    
        for feature in features_notarget:
            ttest_dict[t][feature] = max(ttest_dict[t][feature].values())
        
    result = pd.DataFrame(ttest_dict)
    
    if len(result.columns) == 1:
        if len(result) == 1:
            result = result.iloc[0,0]
        else:
            result = result[result.columns[0]].copy()
            result.sort_values(ascending = False)
    else:
        order = [x for x in result.columns if x in features] + [x for x in features if x not in result.columns]
        result = result.reindex(order)
        
    return result


#!/usr/bin/env python
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright (c) 2015, IBM Corp.
# All rights reserved.
#
# Distributed under the terms of the BSD Simplified License.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from future import standard_library
standard_library.install_aliases()
from collections import OrderedDict

import nzpyida
from nzpyida.internals import idadf_state
from nzpyida.utils import timed, chunklist

import pandas as pd

from nzpyida.feature_selection.private import _check_input


@idadf_state
@timed
def pearson(idadf, target=None, features=None, ignore_indexer=True):
    """
    Compute the pearson correlation coefficients between a set of features and a 
    set of target in an IdaDataFrame. Provide more granualirity than 
    IdaDataFrame.corr
    
    Parameters
    ----------
    idadf : IdaDataFrame
    
    target : str or list of str, optional
        A column or list of columns against to be used as target. Per default, 
        consider all columns
    
    features : str or list of str, optional
        A column or list of columns to be used as features. Per default, 
        consider all columns. 
        
    ignore_indexer : bool, default: True
        Per default, ignore the column declared as indexer in idadf
        
    Returns
    -------
    Pandas.DataFrame or Pandas.Series if only one target
    
    Notes
    -----
    Input columns as target and features should be numerical. 
    
    Examples
    --------
    >>> idadf = IdaDataFrame(idadb, "IRIS")
    >>> pearson(idadf)
    """
    # check if the corr function is installed on the Netezza system
    if idadf._idadb._is_netezza_system():
        corr_query = "SELECT count(*) from _V_OBJECT  where OBJNAME like 'CORR#%' AND OBJDB = CURRENT_DB"
        if idadf._idadb.ida_scalar_query(corr_query) == 0:
            raise NotImplementedError("The CORR function is not installed on the Netezza database.")


    numerical_columns = idadf._get_numerical_columns()
    if features is None:
        features = numerical_columns
        
    target, features = _check_input(idadf, target, features, ignore_indexer)
    
    value_dict = OrderedDict()
    
    for feature in features:
        if feature not in numerical_columns:
            raise TypeError("Correlation-based measure not available for non-numerical column %s"%feature)
                    
    if target == features:
        return idadf.corr(features = features, ignore_indexer=ignore_indexer)
    else:
        for t in target:
            if feature not in numerical_columns:
                raise TypeError("Correlation-based measure not available for non-numerical column %s"%t)
        
        for t in target:
            value_dict[t] = OrderedDict()
            
            features_notarget = [x for x in features if x != t]

            if idadf._idadb._is_netezza_system():
                corr_funct = "CORR"
            else:
                corr_funct = "CORRELATION"

            if len(features_notarget) < 64:
                agg_list = ["%s(\"%s\",\"%s\")"%(corr_funct, x, t) for x in features_notarget]
                agg_string = ', '.join(agg_list)
                name = idadf.internal_state.current_state
                data = idadf.ida_query("SELECT %s FROM %s"%(agg_string, name), first_row_only = True)
            else:
                chunkgen = chunklist(features_notarget, 100)
                data = ()
                for chunk in chunkgen: 
                    agg_list = ["%s(\"%s\",\"%s\")"%(corr_funct, x, t) for x in chunk]
                    agg_string = ', '.join(agg_list)
            
                    name = idadf.internal_state.current_state
                    data += idadf.ida_query("SELECT %s FROM %s"%(agg_string, name), first_row_only = True)
    
            for i, feature in enumerate(features_notarget):
                value_dict[t][feature] = data[i]
        
        ### Fill the matrix
        result = pd.DataFrame(value_dict).fillna(1)
        
        if len(result.columns) == 1:
            if len(result) == 1:
                result = result.iloc[0,0]
            else:
                result = result[result.columns[0]].copy()
                result.sort_values(inplace=True, ascending=False)
        else:
            order = [x for x in result.columns if x in features] + [x for x in features if x not in result.columns]
            result = result.reindex(order)
        
        return result 
  
@idadf_state
@timed          
def spearman(idadf, target=None, features = None, ignore_indexer=True):
    """
    Compute the spearman rho correlation coefficients between a set of features 
    and a set of target in an IdaDataFrame.
    
    Parameters
    ----------
    idadf : IdaDataFrame
    
    target : str or list of str, optional
        A column or list of columns against to be used as target. Per default, 
        consider all columns
    
    features : str or list of str, optional
        A column or list of columns to be used as features. Per default, 
        consider all columns. 
        
    ignore_indexer : bool, default: True
        Per default, ignore the column declared as indexer in idadf
        
    Returns
    -------
    Pandas.DataFrame or Pandas.Series if only one target
    
    Notes
    -----
    Input columns as target and features should be numerical. 
    This function is a wrapper for pearson. 
    The scalability of this approach is not very good. Should not be used on 
    high dimensional data. 
    
    Examples
    --------
    >>> idadf = IdaDataFrame(idadb, "IRIS")
    >>> spearman(idadf)
    """
    numerical_columns = idadf._get_numerical_columns()
    if features is None:
        features = numerical_columns
        
    target, features = _check_input(idadf, target, features, ignore_indexer)
    
    for feature in features:
        if feature not in numerical_columns:
            raise TypeError("Correlation-based measure not available for non-numerical column %s"%feature)
    
    if ignore_indexer is True:
        if idadf.indexer:
            if idadf.indexer in numerical_columns:
                features.remove(idadf.indexer)
    
    if features is None:
        features = list(idadf.columns)
    
    numerical_features = [x for x in features if x in numerical_columns]
    numerical_targets = [x for x in target if x in numerical_columns]
    
    numerical_features = list(set(numerical_features) | set(numerical_targets))
    
    
    agg_list = ["CAST(RANK() OVER (ORDER BY \"%s\") AS INTEGER) AS \"%s_RANK\""%(x, x) for x in numerical_features]
    agg_string = ', '.join(agg_list)
    subselect_stmt = "SELECT %s FROM %s"%(agg_string, idadf.name)

    select_list = ["\"%s_RANK\" AS \"%s\""%(x, x) for x in numerical_features]
    select_string = ', '.join(select_list)
    select_stmt = "SELECT " + select_string + " FROM ( " + subselect_stmt + ") AS T"

    viewname = idadf._idadb._create_view_from_expression(select_stmt)
    
    try:
        idadf_rank = nzpyida.IdaDataFrame(idadf._idadb, viewname)
        return pearson(idadf_rank, target = target, features=numerical_features, ignore_indexer=ignore_indexer)
    except:
        raise
    finally:
        idadf._idadb.drop_view(viewname)
    
    
    
    
 
        
        

        


import pandas as pd
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt
from nzpyida import IdaDataBase, IdaDataFrame
from nzpyida.ae import NZFunTApply
from nzpyida.ae import NZFunApply
from nzpyida.ae import NZFunGroupedApply


#connect to database table
dsn = 'bank'

idadb = IdaDataBase(dsn, 'admin', 'password')

idadf = IdaDataFrame(idadb, 'stocks')

print(idadf.head())

#preprocess data - add more features
idadb.drop_table("stocks_indicator_features")
code_str_host_spus = """def all_stocks_add_features(self,df):







        imputed_df = df.copy()
        imputed_df['DATE'] = pd.to_datetime(imputed_df['DATE'])
        imputed_df = imputed_df.sort_values(by='DATE') 
        imputed_df['DATE'] = imputed_df['DATE'].dt.date



        #name = imputed_df.TICKER[0]



        #how many days in future you need to predict
        future_days = -1

        #add the future close price column and shift by the required days

        imputed_df['FUTURE_CLOSE_PRICE'] = imputed_df['ADJCLOSE'].shift(future_days)



        # add technical indicators
        for n in [14,30,50,200]:

          # create the moving average indicator 
          imputed_df['MA'+str(n)] = imputed_df['ADJCLOSE'].rolling(window=n).mean()


        #imputed_df.dropna(inplace=True)

        def print_output(x):
                row = [x['ID'], x['FUTURE_CLOSE_PRICE'], x['MA14'], x['MA30'], x['MA50'], x['MA200']]
                self.output(row)


        imputed_df.apply(print_output, axis=1)








"""

output_signature = {'ID': 'int', 'FUTURE_CLOSE_PRICE': 'float', 'MA14': 'float', 'MA30': 'float', 'MA50': 'float',
                    'MA200': 'float'}

import time

start = time.time()

nz_tapply = NZFunGroupedApply(df=idadf, code_str=code_str_host_spus, fun_name='all_stocks_add_features', index='TICKER',
                              output_table='stocks_indicator_features', output_signature=output_signature,
                              merge_output_with_df=True)
result = nz_tapply.get_result()
print(result.head())

end = time.time()
print(end - start)


#compute train data sequentially

idadf = IdaDataFrame(idadb, 'stocks_indicator_features')
print(idadf.shape)
idadb.drop_table("stocks_features_train")
code_str_host_spus = """def stocks_train_data(self,df):



        imputed_df = df.copy()
        imputed_df['DATE'] = pd.to_datetime(imputed_df['DATE'])
        imputed_df = imputed_df.sort_values(by='DATE') 
        imputed_df['DATE'] = imputed_df['DATE'].dt.date



        #name = imputed_df.TICKER[0]
        train_size = int(0.9*imputed_df.shape[0])



        train_data = imputed_df[0:train_size]



        def print_output(x):
                row = [x['ID']]
                self.output(row)


        train_data.apply(print_output, axis=1)








"""

output_signature = {'ID': 'int'}

import time

start = time.time()

nz_tapply = NZFunGroupedApply(df=idadf, code_str=code_str_host_spus, fun_name='stocks_train_data', index='TICKER',
                              output_table='stocks_features_train', output_signature=output_signature,
                              merge_output_with_df=True)
result = nz_tapply.get_result()
print(result.head())

end = time.time()
print(end - start)

#compute test data sequentially
idadf = IdaDataFrame(idadb, 'stocks_indicator_features')
print(idadf.shape)

if (idadb.exists_table("stocks_features_test")):
    idadb.drop_table("stocks_features_test")
code_str_host_spus = """def stocks_test_data(self,df):







        imputed_df = df.copy()
        imputed_df['DATE'] = pd.to_datetime(imputed_df['DATE'])
        imputed_df = imputed_df.sort_values(by='DATE') 
        imputed_df['DATE'] = imputed_df['DATE'].dt.date



        #name = imputed_df.TICKER[0]
        train_size = int(0.9*imputed_df.shape[0])



        test_data = imputed_df[train_size:]



        def print_output(x):
                row = [x['ID']]
                self.output(row)


        test_data.apply(print_output, axis=1)








"""

output_signature = {'ID': 'int'}

import time

start = time.time()

nz_tapply = NZFunGroupedApply(df=idadf, code_str=code_str_host_spus, fun_name='stocks_test_data', index='TICKER',
                              output_table='stocks_features_test', output_signature=output_signature,
                              merge_output_with_df=True)
result = nz_tapply.get_result()
print(result.head())

end = time.time()
print(end - start)

#build models on nps with train dataset and save to wml

idadf = IdaDataFrame(idadb, 'stocks_features_train')

code_str_host_spus = """def stocks_rf_ml(self, df):

    import numpy as np

    from sklearn.impute import SimpleImputer
    from sklearn.metrics import mean_squared_error
    from sklearn.ensemble import RandomForestRegressor
    from ibm_watson_machine_learning import APIClient
    from sklearn.model_selection import train_test_split

    wml_credentials = {
                   'url': 'https://us-south.ml.cloud.ibm.com',
                   'apikey':'xxx'
                  }

    client = APIClient(wml_credentials)
    client.set.default_space('xxx')

    imputed_df = df.copy()

    imputed_df['DATE'] = pd.to_datetime(imputed_df['DATE'])
    imputed_df = imputed_df.sort_values(by='DATE')

    imputed_df['DATE']=imputed_df['DATE'].dt.date
    name = imputed_df.TICKER[0]

    from sklearn.preprocessing import LabelEncoder

    temp_dict = dict()





    imputed_df.dropna(inplace=True)


    columns = imputed_df.columns
    for column in columns:

        if column=='ID':
            continue

        #impute missing values 
        # mean for numerical and 'missing' for categorical
        if (imputed_df[column].dtype == 'float64' or imputed_df[column].dtype == 'int64'):
           if imputed_df[column].isnull().sum()==len(imputed_df):

                imputed_df[column] = imputed_df[column].fillna(0)


           else :

                imp = SimpleImputer(missing_values=np.nan, strategy='mean')

                transformed_column = imp.fit_transform(imputed_df[column].values.reshape(-1, 1))

                imputed_df[column] = transformed_column


        if (imputed_df[column].dtype == 'object'):
            # impute missing values for categorical variables


            imp = SimpleImputer(missing_values=None, strategy='constant', fill_value='missing')
            imputed_df[column] = imp.fit_transform(imputed_df[column].values.reshape(-1, 1))
            imputed_df[column] = imputed_df[column].astype('str')

            le = LabelEncoder()

            le.fit(imputed_df[column])
                    # print(le.classes_)
            imputed_df[column] = le.transform(imputed_df[column])
            temp_dict[column] = le

    # Create a random forest regressor
    rf = RandomForestRegressor(n_estimators=200)

    X = imputed_df.drop(['FUTURE_CLOSE_PRICE'], axis=1)
    y = imputed_df['FUTURE_CLOSE_PRICE']


    rf.fit(X,y)

    sw_spec_id = client.software_specifications.get_id_by_name('default_py3.7')
    metadata = {

       client.repository.ModelMetaNames.NAME: 'rf_sklearn_wml_'+name,
       client.repository.ModelMetaNames.SOFTWARE_SPEC_UID: sw_spec_id,
       client.repository.ModelMetaNames.TYPE: 'scikit-learn_0.23'
       }
    model_dic = client.repository.store_model(rf, meta_props=metadata)
    metadata_dic = model_dic.get('metadata')
    model_id = metadata_dic.get('id')
    model_name = metadata_dic.get('name')
    self.output([model_id, model_name])

"""

output_signature = {'model_id': 'str', 'model_name': 'str'}

import time

start = time.time()

nz_groupapply = NZFunGroupedApply(df=idadf, code_str=code_str_host_spus, index='TICKER', fun_name="stocks_rf_ml",
                                  output_signature=output_signature, merge_output_with_df=False)

result = nz_groupapply.get_result()
result = result.as_dataframe()

print(result)
end = time.time()
print(end - start)

#retrieve models from wlm and score the test dataset on nps

idadf = IdaDataFrame(idadb, 'stocks_features_test')
print(idadf.head())

code_str_host_spus = """def stocks_rf_ml(self, df):

    import numpy as np


    from sklearn.impute import SimpleImputer
    from sklearn.metrics import mean_squared_error
    from sklearn.ensemble import RandomForestRegressor
    from ibm_watson_machine_learning import APIClient



    imputed_df = df.copy()
    imputed_df.dropna(inplace=True)
    name = imputed_df.TICKER[0]
    #test = str(self.getDatasliceId())+'_'+str(self.getNumberOfSpus())+'_'+str(self.getNumberOfDataSlices())

    from sklearn.preprocessing import LabelEncoder

    temp_dict = dict()



    #imputed_df.dropna(inplace=True)


    columns = imputed_df.columns
    for column in columns:

        if column=='ID':
            continue

        #impute missing values 
        # mean for numerical and 'missing' for categorical
        if (imputed_df[column].dtype == 'float64' or imputed_df[column].dtype == 'int64'):
           if imputed_df[column].isnull().sum()==len(imputed_df):

                imputed_df[column] = imputed_df[column].fillna(0)


           else :

                imp = SimpleImputer(missing_values=np.nan, strategy='mean')

                transformed_column = imp.fit_transform(imputed_df[column].values.reshape(-1, 1))

                imputed_df[column] = transformed_column


        if (imputed_df[column].dtype == 'object'):
            # impute missing values for categorical variables


            imp = SimpleImputer(missing_values=None, strategy='constant', fill_value='missing')
            imputed_df[column] = imp.fit_transform(imputed_df[column].values.reshape(-1, 1))
            imputed_df[column] = imputed_df[column].astype('str')

            le = LabelEncoder()

            le.fit(imputed_df[column])
                    # print(le.classes_)
            imputed_df[column] = le.transform(imputed_df[column])
            temp_dict[column] = le

    wml_credentials = {
                   'url': 'https://us-south.ml.cloud.ibm.com',
                   'apikey':'xxx'
                  }

    client = APIClient(wml_credentials)
    client.set.default_space('xxx')
    model_dict= client.repository.get_model_details()




    rf=None


    for key, value in model_dict.items():

        if (key=='resources'):
            for res_element in value:
                metadata_dic = res_element.get('metadata')
                #print(metadata_dic)
                model_name ='rf_sklearn_wml_'+name
                if (metadata_dic.get('name')==model_name):
                   rf = client.repository.load(metadata_dic.get('id'))






    X_test = imputed_df.drop(['FUTURE_CLOSE_PRICE'], axis=1)
    y_test = imputed_df['FUTURE_CLOSE_PRICE']

    test_results_df = X_test.copy()

    y_pred = rf.predict(X_test)
    test_results_df['FUTURE_CLOSE_PRICE_PRED'] =y_pred
    test_results_df['FUTURE_CLOSE_PRICE'] =y_test



    accuracy  = rf.score(X_test, y_test)    


    rms = mean_squared_error(y_test, y_pred)

    ds_size = len(X_test)
    test_results_df['DATASET_SIZE'] = ds_size
    test_results_df['ACCURACY']=round(accuracy,2)
    test_results_df['MEAN_SQUARE_ERROR']=round(rms)


    #for all the columns that had label encoders, do an inverse transform

    original_columns = test_results_df.columns

    for column in original_columns:

     if column in temp_dict:   
      test_results_df[column] = temp_dict[column].inverse_transform(test_results_df[column])








    def print_output(x):
                row = [int(x['ID']), x['FUTURE_CLOSE_PRICE_PRED'], x['DATASET_SIZE'], x['ACCURACY']]


                self.output(row)


    test_results_df.apply(print_output, axis=1)



"""

output_signature = {'ID': 'int', 'FUTURE_CLOSE_PRICE_PRED': 'double',
                    'DATASET_SIZE': 'int', 'ACCURACY': 'float'}

import time

start = time.time()

nz_groupapply = NZFunGroupedApply(df=idadf, code_str=code_str_host_spus, index='TICKER', fun_name="stocks_rf_ml",
                                  output_signature=output_signature, merge_output_with_df=True)

result = nz_groupapply.get_result()
result = result.as_dataframe()


print(result)
end = time.time()
print(end - start)




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#-----------------------------------------------------------------------------
# Copyright (c) 2015, IBM Corp.
# All rights reserved.
#
# Distributed under the terms of the BSD Simplified License.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from future import standard_library
standard_library.install_aliases()

from .correlation import  pearson, spearman

from .entropy import entropy, entropy_stats

from .info_gain import info_gain

from .gain_ratio import gain_ratio

from .symmetric_uncertainty import su

from .gini import gini, gini_pairwise
from .discretize import discretize

from .chisquared import chisquared
from .tstats import ttest 

